#!/usr/bin/env python
# -*- coding: utf-8 -*-
__project__ = 'ACAT'
__description__ = 'Alloy Catalysis Automated Toolkit'
__authors__ = ['Shuang Han']
__copyright__ = '2021'
__email__ = 'hanshuangshiren@gmail.com'
__license__ = 'GNU Lesser General Public License 3.0 (LGPL 3.0)'
__maintainer__ = 'Shuang Han'
__url__ = 'https://gitlab.com/asm-dtu/acat'
__version__ = '1.5.2'




from acat.adsorption_sites import ClusterAdsorptionSites
from ase.cluster import Octahedron

atoms = Octahedron('Ni', length=7, cutoff=2)
for atom in atoms:
    if atom.index % 2 == 0:
        atom.symbol = 'Pt' 
atoms.center(vacuum=5.)
cas = ClusterAdsorptionSites(atoms, allow_6fold=False,
                             composition_effect=True,
                             label_sites=True,
                             surrogate_metal='Ni')
site = cas.get_sites()
assert len(site) == 674




from acat.adsorption_sites import SlabAdsorptionSites
from ase.build import fcc211

# Initialize a pure Cu(211) slab
atoms = fcc211('Cu', (3, 3, 4), vacuum=5.)

# Generate a CuAu(211) slab by changing some Cu to Au 
for atom in atoms:
    if atom.index % 2 == 0:
        atom.symbol = 'Au'

# Identify all adsorption sites on the alloy slab
sas = SlabAdsorptionSites(atoms, surface='fcc211',
                          composition_effect=True,
                          label_sites=True,
                          surrogate_metal='Cu')

# Print each site
sites = sas.get_sites()
for i, site in enumerate(sites):
    print('Site {0}: {1}'.format(i, site))

# Print each symmetry-inequivalent site
unique_sites = sas.get_unique_sites()
for i, unique_site in enumerate(unique_sites):
    print('Symmetry-inequivalent site {0}: {1}'.format(i, unique_site))



from acat.adsorption_sites import ClusterAdsorptionSites
from acat.build.adlayer import RandomPatternGenerator as RPG
from acat.build.ordering import RandomOrderingGenerator as ROG
from ase.cluster import Octahedron
from ase.io import read
from ase.visualize import view

# Initialize a pure Ni truncated octahedral nanoparticle 
atoms = Octahedron('Ni', length=5, cutoff=2)
atoms.center(vacuum=5.)

# Generate 50 random adlayer patterns on the nanoparticle with 2 
# of C, O, N adsorbates and a minimum adsorbate-adsorbate distance 
# of 3 Angstrom and save in a trajectory file 'patterns.traj'
rpg = RPG(atoms, adsorbate_species=['C','O','N'], 
          min_adsorbate_distance=3.)
rpg.run(num_gen=50, action='add', num_act=2)

# Visualize the output structures
images = read('patterns.traj', index=':')
view(images)

# If you want to generate random adlayer pattern at a certain coverage,
# try using acat.build.adlayer.max_dist_coverage_pattern. The generated 
# patterns are random but tend to maximize the adsorbate-adsorbate 
# distances so that the generated patterns are more reasonable
from acat.build.adlayer import max_dist_coverage_pattern as maxdcp

# Generate random 0.33 ML CO coverage pattern
pattern = maxdcp(atoms, adsorbate_species='CO', coverage=0.33)
view(pattern)

# Generate random 0.66 ML CO coverage pattern
pattern = maxdcp(atoms, adsorbate_species='CO', coverage=0.66)
view(pattern)

# Another strategy is to provide the minimum adsorbate-adsorbate distance
# and try to maximize the density of the adsorbates
from acat.build.adlayer import min_dist_coverage_pattern as mindcp

# Generate random CO coverage pattern with minimum distance of 2 Angstrom
pattern = mindcp(atoms, adsorbate_species='CO', min_adsorbate_distance=2.)
view(pattern)

# Generate random CO coverage pattern with minimum distance of 5 Angstrom
pattern = mindcp(atoms, adsorbate_species='CO', min_adsorbate_distance=5.)
view(pattern)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Procreation operators meant to be used in symmetry-constrained
genetic algorithm (SCGA)."""
from ..settings import site_heights
from ..utilities import (custom_warning, 
                         is_list_or_tuple, 
                         get_depth)
from ..adsorption_sites import (ClusterAdsorptionSites,    
                                SlabAdsorptionSites)
from ..adsorbate_coverage import (ClusterAdsorbateCoverage,
                                  SlabAdsorbateCoverage)
from ..build.action import (add_adsorbate_to_site, 
                            remove_adsorbates_from_sites)
from ase.ga.offspring_creator import OffspringCreator
from collections import defaultdict
from operator import attrgetter
from copy import deepcopy
import warnings
import random
warnings.formatwarning = custom_warning


class Mutation(OffspringCreator):
    """Base class for all particle mutation type operators.
    Do not call this class directly."""

    def __init__(self, num_muts=1):
        OffspringCreator.__init__(self, num_muts=num_muts)
        self.descriptor = 'Mutation'
        self.min_inputs = 1


class GroupSubstitute(Mutation):
    """Substitute all the atoms in a group with a different element. 
    The elemental composition cannot be fixed.

    Parameters
    ----------
    groups : list of lists or list of list of lists, default None
        The atom indices in each user-divided group. Can be obtained 
        by `acat.build.ordering.SymmetricClusterOrderingGenerator` 
        or `acat.build.ordering.SymmetricSlabOrderingGenerator`.
        You can also mix structures with different groupings in one 
        GA run by providing all possible groups in a list of list of 
        lists, so that the algorithm will randomly assign a grouping
        to the structure, where for each grouping the atoms in each 
        group are of the same type. If not provided here, please 
        provide the groups in atoms.info['data']['groups'] in all 
        intial structures. 

    elements : list of strs, default None
        Only take into account the elements specified in this list. 
        Default is to take all elements into account.

    num_muts : int, default 1
        The number of times to perform this operation.

    """

    def __init__(self, groups=None, 
                 elements=None,
                 num_muts=1):
        Mutation.__init__(self, num_muts=num_muts)

        self.descriptor = 'GroupSubstitute'
        self.elements = elements
        self.groups = groups

    def substitute(self, atoms):
        """Does the actual substitution"""

        atoms = atoms.copy() 
        if self.groups is None:
            assert 'data' in atoms.info 
            assert 'groups' in atoms.info['data']
            groups = atoms.info['data']['groups']
        else:
            depth = get_depth(self.groups)
            if depth > 2:
                random.shuffle(self.groups)
                found = False
                for gs in self.groups:  
                    if all(len(set([atoms[i].symbol for i in g])) == 1 for g in gs):
                        groups = gs
                        found = True
                        break
                if not found:
                    return None                 
            else:
                groups = self.groups
        if self.elements is None:
            e = list(set(atoms.get_chemical_symbols()))
        else:
            e = self.elements

        sorted_elems = sorted(set(atoms.get_chemical_symbols()))
        if e is not None and sorted(e) != sorted_elems:
            for group in groups:
                torem = []
                for i in group:
                    if atoms[i].symbol not in e:
                        torem.append(i)
                for i in torem:
                    group.remove(i)

        itbms = random.sample(range(len(groups)), self.num_muts)
        
        for itbm in itbms:
            mut_group = groups[itbm]
            other_elements = [e for e in self.elements if 
                              e != atoms[mut_group[0]].symbol]
            to_element = random.choice(other_elements)
            atoms.symbols[mut_group] = len(mut_group) * to_element

        return atoms

    def get_new_individual(self, parents):
        f = parents[0]

        indi = self.substitute(f)
        if indi is None:
            return None, '{0} not possible in {1} due to broken symmetry'.format( 
                          self.descriptor, f.info['confid'])                     

        indi = self.initialize_individual(f, indi)
        if 'groups' in f.info['data']:
            indi.info['data']['groups'] = f.info['data']['groups']
        indi.info['data']['parents'] = [f.info['confid']]

        return (self.finalize_individual(indi),
                self.descriptor + ':Parent {0}'.format(f.info['confid']))


class GroupPermutation(Mutation):
    """Permutes the elements in two random groups. The elemental 
    composition can be fixed.

    Parameters
    ----------
    groups : list of lists or list of list of lists, default None
        The atom indices in each user-divided group. Can be obtained 
        by `acat.build.ordering.SymmetricClusterOrderingGenerator` 
        or `acat.build.ordering.SymmetricSlabOrderingGenerator`.
        You can also mix structures with different groupings in one 
        GA run by providing all possible groups in a list of list of
        lists, so that the algorithm will randomly assign a grouping
        to the structure, where for each grouping the atoms in each 
        group are of the same type. If not provided here, please 
        provide the groups in atoms.info['data']['groups'] in all 
        intial structures. 

    elements : list of strs, default None
        Only take into account the elements specified in this list. 
        Default is to take all elements into account.

    keep_composition : bool, defulat False
        Whether the elemental composition should be the same as in
        the parents.

    num_muts : int, default 1
        The number of times to perform this operation.

    """

    def __init__(self, groups=None,
                 elements=None,
                 keep_composition=False, 
                 num_muts=1):
        Mutation.__init__(self, num_muts=num_muts)

        self.descriptor = 'GroupPermutation'
        self.elements = elements
        self.keep_composition = keep_composition
        self.groups = groups

    def get_new_individual(self, parents):

        f = parents[0].copy()
        diffatoms = len(set(f.numbers))

        # Return None if there is only one atomic type
        if diffatoms == 1:
            return None, '{0} not possible in {1}'.format(self.descriptor,
                                                          f.info['confid'])
        indi = self.initialize_individual(f)
        indi.info['data']['parents'] = [f.info['confid']]
        if self.groups is None:
            assert 'data' in f.info
            assert 'groups' in f.info['data']
            groups = f.info['data']['groups']
            indi.info['data']['groups'] = groups
        else:
            depth = get_depth(self.groups)
            if depth > 2:
                random.shuffle(self.groups)
                found = False
                for gs in self.groups:  
                    if all(len(set([f[i].symbol for i in g])) == 1 for g in gs):
                        groups = gs
                        found = True
                        break
                if not found:
                    return None, '{0} not possible in {1} due to broken symmetry'.format(
                                  self.descriptor, f.info['confid'])                     
            else:
                groups = self.groups
            if 'data' in f.info:
                if 'groups' in f.info['data']:
                    indi.info['data']['groups'] = f.info['data']['groups']
        for _ in range(self.num_muts):
            GroupPermutation.mutate(f, groups, self.elements,
                                    self.keep_composition)
        for atom in f:
            indi.append(atom)

        return (self.finalize_individual(indi),
                self.descriptor + ':Parent {0}'.format(f.info['confid']))

    @classmethod
    def mutate(cls, atoms, groups, elements=None, keep_composition=False):
        """Do the actual permutation."""

        if elements is None:
            e = list(set(atoms.get_chemical_symbols()))
        else:
            e = elements

        sorted_elems = sorted(set(atoms.get_chemical_symbols()))
        if e is not None and sorted(e) != sorted_elems:
            for group in groups:
                torem = []
                for i in group:
                    if atoms[i].symbol not in e:
                        torem.append(i)
                for i in torem:
                    group.remove(i)

        if keep_composition:
            dd = defaultdict(list)
            for gi, group in enumerate(groups):
                dd[len(group)].append(gi)
            items = list(dd.items())
            random.shuffle(items)
            mut_gis = None
            for k, v in items:
                if len(v) > 1:
                    mut_gis = v
                    break
            if mut_gis is None:
                return
            random.shuffle(mut_gis)
            i1 = mut_gis[0]
            mut_group1 = groups[i1]           
            options = [i for i in mut_gis[1:] if atoms[mut_group1[0]].symbol 
                       != atoms[groups[i][0]].symbol]

        else:
            i1 = random.randint(0, len(groups) - 1)
            mut_group1 = groups[i1]
            options = [i for i in range(0, len(groups)) if atoms[mut_group1[0]].symbol 
                       != atoms[groups[i][0]].symbol]
        if not options:
            return

        i2 = random.choice(options)
        mut_group2 = groups[i2]
        atoms.symbols[mut_group1+mut_group2] = len(mut_group1) * atoms[
        mut_group2[0]].symbol + len(mut_group2) * atoms[mut_group1[0]].symbol            


class AdsorbateGroupSubstitute(Mutation):
    """Substitute all the adsorbates (or vacancies) in a site group 
    with a different element.

    Parameters
    ----------
    adsorbate_species : str or list of strs 
        A list of possible adsorbate species to be added to the surface.

    species_probabilities : dict, default None                         
        A dictionary that contains keys of each adsorbate species and 
        values of their probabilities of replacing an adsorbate on the 
        surface. Choosing adsorbate species with equal probability if 
        not specified.

    site_groups : list of lists or list of list of lists, default None
        The site indices in each user-divided group. Can be obtained 
        by `acat.build.ordering.SymmetricPatternGenerator`.
        You can also mix structures with different groupings in one 
        GA run by providing all possible groups in a list of list of
        lists, so that the algorithm will randomly assign a grouping
        to the structure, where for each grouping the adsorbate / 
        vacancy occupations in each site group are of the same type. 
        If not provided here, please provide the groups in 
        atoms.info['data']['groups'] in all intial structures. 

    max_species : int, default None
        The maximum allowed adsorbate species (excluding vacancies) for a
        single structure. Allow all adsorbatae species if not specified.

    heights : dict, default acat.settings.site_heights                
        A dictionary that contains the adsorbate height for each site 
        type. Use the default height settings if the height for a site
        type is not specified.

    adsorption_sites : acat.adsorption_sites.ClusterAdsorptionSites object 
        or acat.adsorption_sites.SlabAdsorptionSites object, default None
        Provide the built-in adsorption sites class to accelerate the 
        pattern generation. Make sure all the structures have the same 
        periodicity and atom indexing. If composition_effect=True, you 
        should only provide adsorption_sites when the surface composition 
        is fixed. If this is not provided, the arguments for identifying
        adsorption sites can still be passed in by **kwargs.

    remove_site_shells : int, default 1                                    
        The neighbor shell number within which the neighbor sites should be
        removed. Remove the 1st neighbor site shell by default. Set to 0 if
        no site should be removed.

    remove_site_radius : float, default None                              
        The radius within which the neighbor sites should be removed. This
        serves as an alternative to remove_site_shells.                   

    subtract_heights : bool, default False
        Whether to subtract the height from the bond length when allocating
        a site to an adsorbate. Default is to allocate the site that is
        closest to the adsorbate's binding atom without subtracting height.
        Useful for ensuring the allocated site for each adsorbate is
        consistent with the site to which the adsorbate was added. 

    num_muts : int, default 1
        The number of times to perform this operation.

    dmax : float, default 2.5
        The maximum bond length (in Angstrom) between an atom and its
        nearest site to be considered as the atom being bound to the site.

    """

    def __init__(self, adsorbate_species,
                 species_probabilities=None,
                 site_groups=None, 
                 max_species=None,
                 heights=site_heights,
                 adsorption_sites=None,
                 remove_site_shells=1,
                 remove_site_radius=None,
                 subtract_heights=False,
                 num_muts=1, dmax=2.5, **kwargs):
        Mutation.__init__(self, num_muts=num_muts)

        self.descriptor = 'AdsorbateGroupSubstitute'
        self.adsorbate_species = adsorbate_species if is_list_or_tuple(
                                 adsorbate_species) else [adsorbate_species] 
        self.species_probabilities = species_probabilities
        if self.species_probabilities is not None:
            assert len(self.species_probabilities.keys()) == len(self.adsorbate_species) 
                                                     
        self.site_groups = site_groups
        if max_species is None:
            self.max_species = None
        else:
            self.max_species = min([max_species, len(set(self.adsorbate_species))])
        self.heights = site_heights 
        for k, v in heights.items():
            self.heights[k] = v
        self.adsorption_sites = adsorption_sites
        self.remove_site_shells = remove_site_shells
        self.remove_site_radius = remove_site_radius
        if subtract_heights:
            self.subtract_heights = self.heights
        else:
            self.subtract_heights = None        
        self.dmax = dmax

        self.kwargs = {'allow_6fold': False, 'composition_effect': False, 
                       'ignore_bridge_sites': True, 'label_sites': False}
        self.kwargs.update(kwargs)
        if self.adsorption_sites is not None:
            for k in self.kwargs.keys():             
                self.kwargs[k] = attrgetter(k)(self.adsorption_sites)
        self.__dict__.update(self.kwargs)                                  

    def substitute(self, atoms):
        """Does the actual substitution of the adsorbates"""

        atoms = atoms.copy() 
        if True in atoms.pbc:                                          
            if self.adsorption_sites is not None:
                sas = deepcopy(self.adsorption_sites)
                sas.update(atoms)
            else:
                sas = SlabAdsorptionSites(atoms, **self.kwargs)
            sac = SlabAdsorbateCoverage(atoms, sas, subtract_heights=
                                        self.subtract_heights, dmax=self.dmax)
        else:
            if self.adsorption_sites is not None:
                sas = deepcopy(self.adsorption_sites)
                sas.update(atoms)
            else:
                sas = ClusterAdsorptionSites(atoms, **self.kwargs)
            sac = ClusterAdsorbateCoverage(atoms, sas, subtract_heights=
                                           self.subtract_heights, dmax=self.dmax) 

        hsl = sac.hetero_site_list
        if self.site_groups is None:
            assert 'data' in atoms.info
            assert 'groups' in atoms.info['data']
            groups = atoms.info['data']['groups']
        else:
            depth = get_depth(self.site_groups)
            if depth > 2:
                random.shuffle(self.site_groups)
                found = False
                for gs in self.site_groups:  
                    if all(len(set([hsl[i]['fragment'] for i in g])) == 1 for g in gs):
                        groups = gs
                        found = True
                        break
                if not found:
                    return None 
            else:
                groups = self.site_groups

        ngroups = len(groups)
        indices = list(range(ngroups))
        if (self.remove_site_shells > 0) or (self.remove_site_radius is not None):
            nsl = sas.get_neighbor_site_list(neighbor_number=self.remove_site_shells,
                                             radius=self.remove_site_radius)
            indexes = indices.copy()
            random.shuffle(indexes)
            itbms = set()
            for j in range(len(indexes)):
                itbms.add(j)
                if hsl[groups[j][0]]['occupied']:
                    inset = {ni for i in groups[j] for ni in nsl[i]}
                    ings = [nj for nj in indexes if not set(groups[nj]).isdisjoint(inset)]
                    itbms = set(itbms) - set(ings)
                if len(itbms) == self.num_muts:
                    break
            itbms = list(itbms) 
        else:
            itbms = random.sample(indices, self.num_muts)
        not_mut = [j for j in indices if j not in itbms]
        random.shuffle(not_mut)
        if self.max_species is None:                      
            options = self.adsorbate_species + ['vacancy']
        else:
            not_mut_specs = {hsl[groups[j][0]]['fragment'] for j in not_mut
                             if hsl[groups[j][0]]['occupied']}
            diff = self.max_species - len(not_mut_specs)
            if diff > 0:
                others = [sp for sp in self.adsorbate_species if sp not in not_mut_specs]
                if self.species_probabilities is None:
                    dspecs = random.sample(others, diff)
                else:
                    probs = [self.species_probabilities[osp] for osp in others]      
                    dspecs = random.choices(k=diff, population=others, weights=probs) 
                options = list(not_mut_specs) + dspecs + ['vacancy']
            elif diff == 0:
                options = list(not_mut_specs) + ['vacancy']
            else:
                options = list(not_mut_specs) + ['vacancy']
                warnings.warn('The number of adsorbate species exceeds ' +
                              'the maximum allowed number.')

        changes = [None] * ngroups                                 
        newvs = set()
        indices = itbms + not_mut
        for idx in indices:
            group = groups[idx]
            st0 = hsl[group[0]]
            if not set(group).isdisjoint(newvs):
                if st0['occupied']:
                    changes[idx] = 'vacancy'
            elif idx in itbms:
                if st0['occupied']:
                    spec = st0['fragment']
                else:
                    spec = 'vacancy'     
                other_specs = [sp for sp in options if sp != spec]
                to_spec = random.choice(other_specs)
                changes[idx] = to_spec
                if to_spec == 'vacancy':
                    newvs.update(group)
                elif (self.remove_site_shells > 0) or (self.remove_site_radius is not None):
                    newvs.update([i for k in group for i in nsl[k]])

        rmsites = [hsl[j] for idx, to_spec in enumerate(changes) 
                   if to_spec is not None for j in groups[idx]]
        remove_adsorbates_from_sites(atoms, sites=rmsites, remove_fragments=True)
        for idx, to_spec in enumerate(changes):
            if to_spec not in [None, 'vacancy']:
                for j in groups[idx]:
                    st = hsl[j]
                    height = self.heights[st['site']]
                    add_adsorbate_to_site(atoms, to_spec, st, height)

        if True in atoms.pbc:
            nsac = SlabAdsorbateCoverage(atoms, sas, subtract_heights=
                                         self.subtract_heights, dmax=self.dmax)
        else:
            nsac = ClusterAdsorbateCoverage(atoms, sas, subtract_heights=
                                            self.subtract_heights, dmax=self.dmax)
        atoms.info['data']['adsorbates'] = [t[0] for t in 
            nsac.get_adsorbates(self.adsorbate_species)]

        return atoms

    def get_new_individual(self, parents):
        f = parents[0]
        indi = self.substitute(f)
        if indi is None:
            return None, '{0} not possible in {1} due to broken symmetry'.format(
                         self.descriptor, f.info['confid'])                 

        indi = self.initialize_individual(f, indi) 
        if 'groups' in f.info['data']:
            indi.info['data']['groups'] = f.info['data']['groups']
        indi.info['data']['parents'] = [f.info['confid']] 

        return (self.finalize_individual(indi),
                self.descriptor + ':Parent {0}'.format(f.info['confid']))


class AdsorbateGroupPermutation(Mutation):
    """Permutes the elements in two random site groups.

    Parameters
    ----------
    adsorbate_species : str or list of strs 
        A list of possible adsorbate species to be added to the surface.

    site_groups : list of lists or list of list of lists, default None
        The site indices in each user-divided group. Can be obtained 
        by `acat.build.ordering.SymmetricPatternGenerator`.
        You can also mix structures with different groupings in one 
        GA run by providing all possible groups in a list of list of
        lists, so that the algorithm will randomly assign a grouping
        to the structure, where for each grouping the adsorbate / 
        vacancy occupations in each site group are of the same type.
        If not provided here, please provide the groups in 
        atoms.info['data']['groups'] in all intial structures. 

    heights : dict, default acat.settings.site_heights                
        A dictionary that contains the adsorbate height for each site 
        type. Use the default height settings if the height for a site
        type is not specified.

    adsorption_sites : acat.adsorption_sites.ClusterAdsorptionSites object 
        or acat.adsorption_sites.SlabAdsorptionSites object, default None
        Provide the built-in adsorption sites class to accelerate the 
        pattern generation. Make sure all the structures have the same 
        periodicity and atom indexing. If composition_effect=True, you 
        should only provide adsorption_sites when the surface composition 
        is fixed. If this is not provided, the arguments for identifying
        adsorption sites can still be passed in by **kwargs.

    remove_site_shells : int, default 1                                    
        The neighbor shell number within which the neighbor sites should be
        removed. Remove the 1st neighbor site shell by default. Set to 0 if
        no site should be removed.

    remove_site_radius : float, default None                              
        The radius within which the neighbor sites should be removed. This
        serves as an alternative to remove_site_shells.                   

    subtract_heights : bool, default False
        Whether to subtract the height from the bond length when allocating
        a site to an adsorbate. Default is to allocate the site that is
        closest to the adsorbate's binding atom without subtracting height.
        Useful for ensuring the allocated site for each adsorbate is
        consistent with the site to which the adsorbate was added. 

    num_muts : int, default 1
        The number of times to perform this operation.

    dmax : float, default 2.5
        The maximum bond length (in Angstrom) between an atom and its
        nearest site to be considered as the atom being bound to the site.

    """

    def __init__(self, adsorbate_species, 
                 site_groups=None,
                 heights=site_heights,
                 adsorption_sites=None,
                 remove_site_shells=1,
                 remove_site_radius=None,
                 subtract_heights=False,
                 num_muts=1, dmax=2.5, **kwargs):
        Mutation.__init__(self, num_muts=num_muts)

        self.descriptor = 'AdsorbateGroupPermutation'
        self.adsorbate_species = adsorbate_species if is_list_or_tuple(       
                                 adsorbate_species) else [adsorbate_species]
        self.site_groups = site_groups
        self.heights = site_heights 
        for k, v in heights.items():
            self.heights[k] = v
        self.adsorption_sites = adsorption_sites
        self.remove_site_shells = remove_site_shells
        self.remove_site_radius = remove_site_radius
        if subtract_heights:
            self.subtract_heights = self.heights
        else:
            self.subtract_heights = None        
        self.dmax = dmax

        self.kwargs = {'allow_6fold': False, 'composition_effect': False, 
                       'ignore_bridge_sites': True, 'label_sites': False}
        self.kwargs.update(kwargs)
        if self.adsorption_sites is not None:
            for k in self.kwargs.keys():             
                self.kwargs[k] = attrgetter(k)(self.adsorption_sites)
        self.__dict__.update(self.kwargs)                                  

    def get_new_individual(self, parents):

        f = parents[0].copy()
        
        indi = self.initialize_individual(f)
        indi.info['data']['parents'] = [f.info['confid']]

        if True in f.pbc:                                          
            if self.adsorption_sites is not None:
                sas = deepcopy(self.adsorption_sites)
                sas.update(f)
            else:
                sas = SlabAdsorptionSites(f, **self.kwargs)
            sac = SlabAdsorbateCoverage(f, sas, subtract_heights=
                                        self.subtract_heights, dmax=self.dmax)
        else:
            if self.adsorption_sites is not None:
                sas = deepcopy(self.adsorption_sites)
                sas.update(f)
            else:
                sas = ClusterAdsorptionSites(f, **self.kwargs)
            sac = ClusterAdsorbateCoverage(f, sas, subtract_heights=
                                           self.subtract_heights, dmax=self.dmax) 
        hsl = sac.hetero_site_list

        if self.site_groups is None:
            assert 'data' in atoms.info
            assert 'groups' in atoms.info['data']
            site_groups = f.info['data']['groups']
            indi.info['data']['groups'] = site_groups
        else:
            depth = get_depth(self.site_groups)
            if depth > 2:
                random.shuffle(self.site_groups)
                found = False
                for gs in self.site_groups:  
                    if all(len(set([hsl[i]['fragment'] for i in g])) == 1 for g in gs):
                        site_groups = gs
                        found = True
                        break
                if not found:
                    return None, '{0} not possible in {1} due to broken symmetry'.format(
                                  self.descriptor, f.info['confid'])                     
            else:
                site_groups = self.site_groups
            if 'groups' in f.info['data']:
                indi.info['data']['groups'] = f.info['data']['groups']

        # Return None if there is only one adsorbate (vacancy) type
        if set([hsl[g[0]]['fragment'] for g in site_groups]) == 1:
            return None, '{0} not possible in {1}'.format(self.descriptor,
                                                          f.info['confid'])
        for _ in range(self.num_muts):
            AdsorbateGroupPermutation.mutate(f, site_groups, self.heights, sas, 
                                             hsl, self.remove_site_shells, 
                                             self.remove_site_radius)
            if True in indi.pbc:
                nsac = SlabAdsorbateCoverage(f, sas, subtract_heights=
                                             self.subtract_heights, dmax=self.dmax)
            else:
                nsac = ClusterAdsorbateCoverage(f, sas, subtract_heights=
                                                self.subtract_heights, dmax=self.dmax)
            hsl = nsac.hetero_site_list

        for atom in f:
            indi.append(atom)
        
        indi.info['data']['adsorbates'] = [t[0] for t in 
            nsac.get_adsorbates(self.adsorbate_species)]

        return (self.finalize_individual(indi),
                self.descriptor + ':Parent {0}'.format(f.info['confid']))

    @classmethod
    def mutate(cls, atoms, groups, heights, adsorption_sites, 
               hetero_site_list, remove_site_shells, remove_site_radius):
        """Do the actual permutation of the adsorbates."""

        sas, hsl = adsorption_sites, hetero_site_list
        ngroups = len(groups)
        indices = list(range(ngroups))
        i1 = random.randint(0, ngroups - 1)
        st01 = hsl[groups[i1][0]]
        ings = []
        if (remove_site_shells > 0) or (remove_site_radius is not None):                                                   
            nsl = sas.get_neighbor_site_list(neighbor_number=remove_site_shells,
                                             radius=remove_site_radius) 
            if st01['occupied']:
                ings += [nj for nj in indices if not set(groups[nj]).isdisjoint(nsl[i1])]
        options = [j for j in range(ngroups) if (j not in ings) and 
                   (st01['fragment'] != hsl[groups[j][0]]['fragment'])]
        if not options:
            return
        i2 = random.choice(options)
        st02 = hsl[groups[i2][0]]
        if st01['occupied']:
            to_spec2 = st01['fragment']
        else: 
            to_spec2 = 'vacancy'
        if st02['occupied']:
            to_spec1 = st02['fragment']
        else:
            to_spec1 = 'vacancy'
        not_mut = [j for j in indices if j not in [i1, i2]]
        random.shuffle(not_mut)

        changes = [None] * ngroups 
        newvs = set()
        indices = [i1, i2] + not_mut
        for idx in indices:
            group = groups[idx]
            st0 = hsl[group[0]]
            if not set(group).isdisjoint(newvs): 
                if st0['occupied']:
                    changes[idx] = 'vacancy'
            elif idx == i1:
                changes[idx] = to_spec1
                if to_spec1 == 'vacancy':
                    newvs.update(group)
                elif (remove_site_shells > 0) or (remove_site_radius is not None):
                    newvs.update([i for k in group for i in nsl[k]])
            elif idx == i2:
                changes[idx] = to_spec2
                if to_spec2 == 'vacancy':
                    newvs.update(group)
                elif (remove_site_shells > 0) or (remove_site_radius is not None):
                    newvs.update([i for k in group for i in nsl[k]])

        rmsites = [hsl[j] for idx, to_spec in enumerate(changes) 
                   if to_spec is not None for j in groups[idx]]
        remove_adsorbates_from_sites(atoms, sites=rmsites, remove_fragments=True)
        for idx, to_spec in enumerate(changes):
            if to_spec not in [None, 'vacancy']:
                for j in groups[idx]:
                    st = hsl[j]
                    height = heights[st['site']]
                    add_adsorbate_to_site(atoms, to_spec, st, height)


class Crossover(OffspringCreator):
    """Base class for all particle crossovers.
    Do not call this class directly."""
    def __init__(self):
        OffspringCreator.__init__(self)
        self.descriptor = 'Crossover'
        self.min_inputs = 2


class GroupCrossover(Crossover):
    """Merge the elemental distributions in two half groups from 
    different structures together. The elemental composition can be 
    fixed.

    Parameters
    ----------
    groups : list of lists or list of list of lists, default None
        The atom indices in each user-divided group. Can be obtained 
        by `acat.build.ordering.SymmetricClusterOrderingGenerator` 
        or `acat.build.ordering.SymmetricSlabOrderingGenerator`.
        You can also mix structures with different groupings in one 
        GA run by providing all possible groups in a list of list of
        lists, so that the algorithm will randomly assign a grouping
        to the structure, where for each grouping the atoms in each 
        group are of the same type. If not provided here, please 
        provide the groups in atoms.info['data']['groups'] in all 
        intial structures. 

    elements : list of strs, default None
        Only take into account the elements specified in this list. 
        Default is to take all elements into account.

    keep_composition : bool, defulat False
        Whether the elemental composition should be the same as in
        the parents.

    """

    def __init__(self, groups=None, elements=None, keep_composition=False):
        Crossover.__init__(self)
        self.groups = groups
        self.elements = elements
        self.keep_composition = keep_composition
        self.descriptor = 'GroupCrossover'
        
    def get_new_individual(self, parents):

        f, m = parents
        indi = f.copy()
        if self.groups is None:
            assert 'data' in f.info
            assert 'groups' in f.info['data']
            assert 'data' in m.info
            assert 'groups' in m.info['data']
            assert f.info['data']['groups'] == m.info['data']['groups']
            groups = indi.info['data']['groups']
        else:
            depth = get_depth(self.groups)
            if depth > 2:
                random.shuffle(self.groups)
                found = False
                for gs in self.groups:  
                    if all(len(set([atoms[i].symbol for i in g])) == 1 for g in gs):
                        groups = gs
                        found = True
                        break
                if not found:
                    return None, '{0} not possible in {1} due to broken symmetry'.format(
                                  self.descriptor, f.info['confid'])                      
            else:
                groups = self.groups.copy()
        if self.elements is None:
            e = list(set(f.get_chemical_symbols()))
        else:
            e = self.elements

        sorted_elems = sorted(set(f.get_chemical_symbols()))
        if e is not None and sorted(e) != sorted_elems:
            for group in groups:
                torem = []
                for i in group:
                    if f[i].symbol not in e:
                        torem.append(i)
                for i in torem:
                    group.remove(i)
        random.shuffle(groups)

        if self.keep_composition:
            def fix_composition_swaps(groups1, groups2):
                indices = sorted([i for j in groups1 for i in j])                                
                zipped = list(map(list, zip(groups1, groups2)))
                gids = [i for i, (groups1, groups2) in enumerate(zipped) 
                        if groups1 != groups2]

                # If solution not found in 1000 iterations, we say there
                # is no possible solution at all
                gids_list = []
                for j in range(1000):        
                    random.shuffle(gids)
                    if gids in gids_list:
                        continue
                    gids_list.append(gids.copy())

                    for n, i in enumerate(gids):
                        zipped[i].reverse()
                        if indices == sorted(idx for groups1, _ in zipped 
                        for idx in groups1):
                            return gids[:n+1]
                    zipped = list(map(list, zip(groups1, groups2)))
                return []

            fsyms = [list(f.symbols[g]) for g in groups]
            msyms = [list(m.symbols[g]) for g in groups]
            swap_ids = fix_composition_swaps(fsyms, msyms)
            mids = [i for j in swap_ids for i in groups[j]] 

        else:
            mgroups = groups[len(groups)//2:]
            mids = [i for group in mgroups for i in group]

        indi.symbols[mids] = m.symbols[mids]
        indi = self.initialize_individual(f, indi)
        if 'data' in f.info:
            if 'groups' in f.info['data']:
                indi.info['data']['groups'] = f.info['data']['groups']
        indi.info['data']['parents'] = [i.info['confid'] for i in parents] 
        indi.info['data']['operation'] = 'crossover'
        parent_message = ':Parents {0} {1}'.format(f.info['confid'],
                                                   m.info['confid']) 

        return (self.finalize_individual(indi),
                self.descriptor + parent_message)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from itertools import product 

# Use this label dictionary when site compostion is 
# not considered. Useful for monometallic clusters.
def get_monometallic_cluster_labels():
    return {'ontop|vertex': 1,
            'ontop|edge': 2,
            'ontop|fcc111': 3,
            'ontop|fcc100': 4,
            'bridge|edge': 5,
            'bridge|fcc111': 6,
            'bridge|fcc100': 7,
            'fcc|fcc111': 8,
            'hcp|fcc111': 9,
            '4fold|fcc100': 10,
            '6fold|subsurf': 11}


# Use this label dictionary when site compostion is 
# considered. Useful for bimetallic clusters.
def get_bimetallic_cluster_labels(metals): 
    ma, mb = metals[0], metals[1]

    return {'ontop|vertex|{}'.format(ma): 1, 
            'ontop|vertex|{}'.format(mb): 2,
            'ontop|edge|{}'.format(ma): 3,
            'ontop|edge|{}'.format(mb): 4,
            'ontop|fcc111|{}'.format(ma): 5,
            'ontop|fcc111|{}'.format(mb): 6,
            'ontop|fcc100|{}'.format(ma): 7,
            'ontop|fcc100|{}'.format(mb): 8,
            'bridge|edge|{}{}'.format(ma,ma): 9, 
            'bridge|edge|{}{}'.format(ma,mb): 10,
            'bridge|edge|{}{}'.format(mb,mb): 11,
            'bridge|fcc111|{}{}'.format(ma,ma): 12,
            'bridge|fcc111|{}{}'.format(ma,mb): 13,
            'bridge|fcc111|{}{}'.format(mb,mb): 14,
            'bridge|fcc100|{}{}'.format(ma,ma): 15,
            'bridge|fcc100|{}{}'.format(ma,mb): 16,
            'bridge|fcc100|{}{}'.format(mb,mb): 17,
            'fcc|fcc111|{}{}{}'.format(ma,ma,ma): 18,
            'fcc|fcc111|{}{}{}'.format(ma,ma,mb): 19, 
            'fcc|fcc111|{}{}{}'.format(ma,mb,mb): 20,
            'fcc|fcc111|{}{}{}'.format(mb,mb,mb): 21,
            'hcp|fcc111|{}{}{}'.format(ma,ma,ma): 22,
            'hcp|fcc111|{}{}{}'.format(ma,ma,mb): 23,
            'hcp|fcc111|{}{}{}'.format(ma,mb,mb): 24,
            'hcp|fcc111|{}{}{}'.format(mb,mb,mb): 25,
            '4fold|fcc100|{}{}{}{}'.format(ma,ma,ma,ma): 26,
            '4fold|fcc100|{}{}{}{}'.format(ma,ma,ma,mb): 27, 
            '4fold|fcc100|{}{}{}{}'.format(ma,ma,mb,mb): 28,
            '4fold|fcc100|{}{}{}{}'.format(ma,mb,ma,mb): 29, 
            '4fold|fcc100|{}{}{}{}'.format(ma,mb,mb,mb): 30,
            '4fold|fcc100|{}{}{}{}'.format(mb,mb,mb,mb): 31,
            '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,ma,ma,ma,ma): 32,
            '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,ma,ma,ma,mb): 33,
            '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,ma,mb,ma,ma): 34,
            '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,ma,ma,mb,mb): 35,
            '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,ma,mb,mb,ma): 36,
            '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,ma,mb,mb,mb): 37,
            '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,mb,ma,ma,ma): 38,
            '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,mb,ma,ma,mb): 39,
            '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,mb,mb,ma,ma): 40,
            '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,mb,ma,mb,mb): 41,
            '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,mb,mb,mb,ma): 42,
            '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,mb,mb,mb,mb): 43,
            '6fold|subsurf|{}{}{}{}{}{}'.format(ma,mb,mb,ma,ma,ma): 44,
            '6fold|subsurf|{}{}{}{}{}{}'.format(ma,mb,mb,ma,ma,mb): 45,
            '6fold|subsurf|{}{}{}{}{}{}'.format(ma,mb,mb,mb,ma,ma): 46,
            '6fold|subsurf|{}{}{}{}{}{}'.format(ma,mb,mb,ma,mb,mb): 47,
            '6fold|subsurf|{}{}{}{}{}{}'.format(ma,mb,mb,mb,mb,ma): 48,
            '6fold|subsurf|{}{}{}{}{}{}'.format(ma,mb,mb,mb,mb,mb): 49,
            '6fold|subsurf|{}{}{}{}{}{}'.format(mb,mb,mb,ma,ma,ma): 50,
            '6fold|subsurf|{}{}{}{}{}{}'.format(mb,mb,mb,ma,ma,mb): 51,
            '6fold|subsurf|{}{}{}{}{}{}'.format(mb,mb,mb,mb,ma,ma): 52,
            '6fold|subsurf|{}{}{}{}{}{}'.format(mb,mb,mb,ma,mb,mb): 53,
            '6fold|subsurf|{}{}{}{}{}{}'.format(mb,mb,mb,mb,mb,ma): 54,
            '6fold|subsurf|{}{}{}{}{}{}'.format(mb,mb,mb,mb,mb,mb): 55}


# Use this label dictionary when site compostion is 
# considered. Useful for clusters with arbitrary number
# of components.
def get_multimetallic_cluster_labels(metals): 
    relations = {'ontop': ['vertex','edge','fcc111','fcc100'],
                 'bridge': ['edge','fcc111','fcc100'],
                 'fcc': ['fcc111'],
                 'hcp': ['fcc111'],
                 '4fold': ['fcc100'],
                 '6fold': ['subsurf']}
    d = {}
    count = 0 
    for k, vs in relations.items():
        if k == 'ontop':
            n = 1
        elif k == 'bridge':
            n = 2
        elif k in ['fcc','hcp']:
            n = 3
        elif k == '4fold':
            n = 4
        elif k == '6fold':
            n = 6
        for v in vs:
            for p in product(metals, repeat=n):
                comp = ''.join(p)
                count += 1
                d['{0}|{1}|{2}'.format(k, v, comp)] = count    
                
    return d


# Use this label dictionary when site compostion is 
# not considered. Useful for monometallic slabs.
def get_monometallic_slab_labels(surface): 
    if surface in ['fcc111','hcp0001']:
        return {'ontop|terrace': 1,
                'bridge|terrace': 2,
                'fcc|terrace': 3,
                'hcp|terrace': 4,
                '6fold|subsurf': 5}

    elif surface in ['fcc100','bcc100']:
        return {'ontop|terrace': 1,
                'bridge|terrace': 2,
                '4fold|terrace': 3}

    elif surface in ['fcc110','hcp10m10h']:
        return {'ontop|step': 1,
                'bridge|step': 2, 
                'bridge|sc-tc-h': 3,
                'fcc|sc-tc-h': 4,
                'hcp|sc-tc-h': 5,
                '4fold|terrace': 6,
                '5fold|terrace': 7,
                '6fold|subsurf': 8}

    elif surface == 'fcc211':
        return {'ontop|step': 1,      
                'ontop|terrace': 2,
                'ontop|corner': 3, 
                'bridge|step': 4,
                'bridge|corner': 5,
                'bridge|sc-tc-h': 6,
                'bridge|tc-cc-h': 7,
                'bridge|sc-cc-t': 8,
                'fcc|sc-tc-h': 9,
                'fcc|tc-cc-h': 10,
                'hcp|sc-tc-h': 11,
                'hcp|tc-cc-h': 12,
                '4fold|sc-cc-t': 13,
                '6fold|subsurf': 14}

    elif surface in ['fcc311','fcc331']:
        return {'ontop|step': 1,
                'ontop|terrace': 2,
                'bridge|step': 3,
                'bridge|terrace': 4,
                'bridge|sc-tc-h': 5,
                'bridge|sc-tc-t': 6,
                'fcc|sc-tc-h': 7,
                'hcp|sc-tc-h': 8,
                '4fold|sc-tc-t': 9,
                '6fold|subsurf': 10}

    elif surface == 'fcc322':
        return {'ontop|step': 1,
                'ontop|terrace': 2,
                'ontop|corner': 3,
                'bridge|step': 4,
                'bridge|terrace': 5,
                'bridge|corner': 6,
                'bridge|sc-tc-h': 7,
                'bridge|tc-cc-h': 8,
                'bridge|sc-cc-t': 9,
                'fcc|terrace': 10,
                'fcc|sc-tc-h': 11,
                'fcc|tc-cc-h': 12,
                'hcp|terrace': 13,
                'hcp|sc-tc-h': 14,                    
                'hcp|tc-cc-h': 15,
                '4fold|sc-cc-t': 16,
                '6fold|subsurf': 17}

    elif surface in ['fcc221','fcc332']:
        return {'ontop|step': 1,
                'ontop|terrace': 2,
                'ontop|corner': 3,
                'bridge|step': 4,
                'bridge|terrace': 5,
                'bridge|corner': 6,
                'bridge|sc-tc-h': 7,
                'bridge|tc-cc-h': 8,
                'bridge|sc-cc-h': 9,
                'fcc|terrace': 10,
                'fcc|sc-tc-h': 11,
                'fcc|tc-cc-h': 12,
                'fcc|sc-cc-h': 13,
                'hcp|terrace': 14,
                'hcp|sc-tc-h': 15, 
                'hcp|tc-cc-h': 16,
                'hcp|sc-cc-h': 17,
                '6fold|subsurf': 18}

    elif surface == 'bcc110':
        return {'ontop|terrace': 1,
                'shortbridge|terrace': 2,
                'longbridge|terrace': 3,
                '3fold|terrace': 4}
              
    elif surface == 'bcc111':           
        return {'ontop|step': 1,                       
                'ontop|terrace': 2,        
                'ontop|corner': 3,
                'shortbridge|sc-tc-o': 4,
                'shortbridge|tc-cc-o': 5,
                'longbridge|sc-cc-o': 6,
                '3fold|sc-tc-cc-o': 7}

    elif surface == 'bcc210':
        return {'ontop|step': 1,     
                'ontop|terrace': 2,
                'ontop|corner': 3, 
                'bridge|step': 4,
                'bridge|terrace': 5,
                'bridge|corner': 6,
                'bridge|sc-tc-o': 7,
                'bridge|tc-cc-o': 8,
                'bridge|sc-cc-t': 9,
                '3fold|sc-tc-o': 10,
                '3fold|tc-cc-o': 11,
                '4fold|sc-cc-t': 12}

    elif surface == 'bcc211':
        return {'ontop|step': 1,
                'bridge|step': 2, 
                'bridge|sc-tc-o': 3,
                '3fold|sc-tc-o': 4,
                '4fold|terrace': 5,
                '5fold|terrace': 6}

    elif surface == 'bcc310':
        return {'ontop|step': 1,
                'ontop|terrace': 2,
                'bridge|step': 3,
                'bridge|terrace': 4,
                'bridge|sc-tc-o': 5,
                'bridge|sc-tc-t': 6,
                '3fold|sc-tc-o': 7,
                '4fold|sc-tc-t': 8}

    elif surface == 'hcp10m10t':
        return {'ontop|step': 1,
                'ontop|terrace': 2,
                'bridge|step': 3,
                'bridge|terrace': 4,
                'bridge|sc-tc-t': 5,
                '5fold|subsurf': 6,}

    elif surface == 'hcp10m11':
        return {'ontop|step': 1,
                'ontop|terrace': 2,
                'bridge|step': 3,
                'bridge|terrace': 4,
                'bridge|sc-tc-h': 5,
                'fcc|sc-tc-h': 6,
                'hcp|sc-tc-h': 7,
                '4fold|subsurf': 8,
                '5fold|subsurf': 9,
                '6fold|subsurf': 10}

    elif surface == 'hcp10m12':       
        return {'ontop|step': 1,
                'ontop|terrace': 2,
                'ontop|corner': 3,
                'bridge|step': 4,
                'bridge|terrace': 5,
                'bridge|corner': 6,
                'bridge|sc-tc-h': 7,
                'bridge|tc-cc-t': 8,
                'bridge|sc-cc-h': 9,
                'fcc|sc-tc-h': 10,
                'fcc|sc-cc-h': 11,
                'hcp|sc-tc-h': 12,
                'hcp|sc-cc-h': 13,
                '4fold|tc-cc-t': 14,
                '6fold|subsurf': 15}


# Use this label dictionary when site compostion is 
# considered. Useful for bimetallic slabs.
def get_bimetallic_slab_labels(surface, metals): 
    ma, mb = metals[0], metals[1]

    if surface in ['fcc111','hcp0001']:
        return {'ontop|terrace|{}'.format(ma): 1, 
                'ontop|terrace|{}'.format(mb): 2,
                'bridge|terrace|{}{}'.format(ma,ma): 3, 
                'bridge|terrace|{}{}'.format(ma,mb): 4,
                'bridge|terrace|{}{}'.format(mb,mb): 5, 
                'fcc|terrace|{}{}{}'.format(ma,ma,ma): 6,
                'fcc|terrace|{}{}{}'.format(ma,ma,mb): 7, 
                'fcc|terrace|{}{}{}'.format(ma,mb,mb): 8,
                'fcc|terrace|{}{}{}'.format(mb,mb,mb): 9,
                'hcp|terrace|{}{}{}'.format(ma,ma,ma): 10,
                'hcp|terrace|{}{}{}'.format(ma,ma,mb): 11,
                'hcp|terrace|{}{}{}'.format(ma,mb,mb): 12,
                'hcp|terrace|{}{}{}'.format(mb,mb,mb): 13,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,ma,ma,ma,ma): 14,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,ma,ma,ma,mb): 15,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,ma,mb,ma,ma): 16,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,ma,ma,mb,mb): 17,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,ma,mb,mb,ma): 18,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,ma,mb,mb,mb): 19,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,mb,ma,ma,ma): 20,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,mb,ma,ma,mb): 21,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,mb,mb,ma,ma): 22,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,mb,ma,mb,mb): 23,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,mb,mb,mb,ma): 24,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,mb,mb,mb,mb): 25,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,mb,mb,ma,ma,ma): 26,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,mb,mb,ma,ma,mb): 27,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,mb,mb,mb,ma,ma): 28,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,mb,mb,ma,mb,mb): 29,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,mb,mb,mb,mb,ma): 30,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,mb,mb,mb,mb,mb): 31,
                '6fold|subsurf|{}{}{}{}{}{}'.format(mb,mb,mb,ma,ma,ma): 32,
                '6fold|subsurf|{}{}{}{}{}{}'.format(mb,mb,mb,ma,ma,mb): 33,
                '6fold|subsurf|{}{}{}{}{}{}'.format(mb,mb,mb,mb,ma,ma): 34,
                '6fold|subsurf|{}{}{}{}{}{}'.format(mb,mb,mb,ma,mb,mb): 35,
                '6fold|subsurf|{}{}{}{}{}{}'.format(mb,mb,mb,mb,mb,ma): 36,
                '6fold|subsurf|{}{}{}{}{}{}'.format(mb,mb,mb,mb,mb,mb): 37}

    elif surface in ['fcc100','bcc100']:
        return {'ontop|terrace|{}'.format(ma): 1, 
                'ontop|terrace|{}'.format(mb): 2,
                'bridge|terrace|{}{}'.format(ma,ma): 3, 
                'bridge|terrace|{}{}'.format(ma,mb): 4,
                'bridge|terrace|{}{}'.format(mb,mb): 5, 
                '4fold|terrace|{}{}{}{}'.format(ma,ma,ma,ma): 6,
                '4fold|terrace|{}{}{}{}'.format(ma,ma,ma,mb): 7, 
                '4fold|terrace|{}{}{}{}'.format(ma,ma,mb,mb): 8,
                '4fold|terrace|{}{}{}{}'.format(ma,mb,ma,mb): 9, 
                '4fold|terrace|{}{}{}{}'.format(ma,mb,mb,mb): 10,
                '4fold|terrace|{}{}{}{}'.format(mb,mb,mb,mb): 11}

    elif surface in ['fcc110','hcp10m10h']:
        return {'ontop|step|{}'.format(ma): 1,
                'ontop|step|{}'.format(mb): 2,
                'bridge|step|{}{}'.format(ma,ma): 3,
                'bridge|step|{}{}'.format(ma,mb): 4,
                'bridge|step|{}{}'.format(mb,mb): 5,
                'bridge|sc-tc-h|{}{}'.format(ma,ma): 6,
                'bridge|sc-tc-h|{}{}'.format(ma,mb): 7,
                'bridge|sc-tc-h|{}{}'.format(mb,mb): 8,
                'fcc|sc-tc-h|{}{}{}'.format(ma,ma,ma): 9,
                'fcc|sc-tc-h|{}{}{}'.format(ma,ma,mb): 10, 
                'fcc|sc-tc-h|{}{}{}'.format(ma,mb,mb): 11,
                'fcc|sc-tc-h|{}{}{}'.format(mb,mb,mb): 12,
                'hcp|sc-tc-h|{}{}{}'.format(ma,ma,ma): 13,
                'hcp|sc-tc-h|{}{}{}'.format(ma,ma,mb): 14,
                'hcp|sc-tc-h|{}{}{}'.format(ma,mb,mb): 15,
                'hcp|sc-tc-h|{}{}{}'.format(mb,mb,mb): 16,
                '4fold|terrace|{}{}-{}{}'.format(ma,ma,ma,ma): 17,
                '4fold|terrace|{}{}-{}{}'.format(ma,ma,ma,mb): 18,
                '4fold|terrace|{}{}-{}{}'.format(ma,ma,mb,mb): 19,
                '4fold|terrace|{}{}-{}{}'.format(ma,mb,ma,ma): 20,
                '4fold|terrace|{}{}-{}{}'.format(ma,mb,ma,mb): 21,
                '4fold|terrace|{}{}-{}{}'.format(ma,mb,mb,mb): 22,
                '4fold|terrace|{}{}-{}{}'.format(mb,mb,ma,ma): 23,
                '4fold|terrace|{}{}-{}{}'.format(mb,mb,ma,mb): 24,
                '4fold|terrace|{}{}-{}{}'.format(mb,mb,mb,mb): 25,
                # neighbor elements count clockwise from shorter bond ma
                '5fold|terrace|{}-{}{}{}{}'.format(ma,ma,ma,ma,ma): 26,
                '5fold|terrace|{}-{}{}{}{}'.format(ma,ma,ma,ma,mb): 27,
                '5fold|terrace|{}-{}{}{}{}'.format(ma,ma,ma,mb,mb): 28,
                '5fold|terrace|{}-{}{}{}{}'.format(ma,ma,mb,ma,mb): 29,
                '5fold|terrace|{}-{}{}{}{}'.format(ma,ma,mb,mb,ma): 30,
                '5fold|terrace|{}-{}{}{}{}'.format(ma,ma,mb,mb,mb): 31,
                '5fold|terrace|{}-{}{}{}{}'.format(ma,mb,mb,mb,mb): 32,
                '5fold|terrace|{}-{}{}{}{}'.format(mb,ma,ma,ma,ma): 33,
                '5fold|terrace|{}-{}{}{}{}'.format(mb,ma,ma,ma,mb): 34,
                '5fold|terrace|{}-{}{}{}{}'.format(mb,ma,ma,mb,mb): 35,
                '5fold|terrace|{}-{}{}{}{}'.format(mb,ma,mb,ma,mb): 36,
                '5fold|terrace|{}-{}{}{}{}'.format(mb,ma,mb,mb,ma): 37,
                '5fold|terrace|{}-{}{}{}{}'.format(mb,ma,mb,mb,mb): 38,
                '5fold|terrace|{}-{}{}{}{}'.format(mb,mb,mb,mb,mb): 39,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,ma,ma,ma,ma): 40,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,ma,ma,ma,mb): 41,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,ma,mb,ma,ma): 42,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,ma,ma,mb,mb): 43,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,ma,mb,mb,ma): 44,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,ma,mb,mb,mb): 45,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,mb,ma,ma,ma): 46,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,mb,ma,ma,mb): 47,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,mb,mb,ma,ma): 48,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,mb,ma,mb,mb): 49,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,mb,mb,mb,ma): 50,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,mb,mb,mb,mb): 51,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,mb,mb,ma,ma,ma): 52,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,mb,mb,ma,ma,mb): 53,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,mb,mb,mb,ma,ma): 54,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,mb,mb,ma,mb,mb): 55,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,mb,mb,mb,mb,ma): 56,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,mb,mb,mb,mb,mb): 57,
                '6fold|subsurf|{}{}{}{}{}{}'.format(mb,mb,mb,ma,ma,ma): 58,
                '6fold|subsurf|{}{}{}{}{}{}'.format(mb,mb,mb,ma,ma,mb): 59,
                '6fold|subsurf|{}{}{}{}{}{}'.format(mb,mb,mb,mb,ma,ma): 60,
                '6fold|subsurf|{}{}{}{}{}{}'.format(mb,mb,mb,ma,mb,mb): 61,
                '6fold|subsurf|{}{}{}{}{}{}'.format(mb,mb,mb,mb,mb,ma): 62,
                '6fold|subsurf|{}{}{}{}{}{}'.format(mb,mb,mb,mb,mb,mb): 63}

    elif surface == 'fcc211':
        return {'ontop|step|{}'.format(ma): 1,
                'ontop|step|{}'.format(mb): 2,
                'ontop|terrace|{}'.format(ma): 3,
                'ontop|terrace|{}'.format(mb): 4,
                'ontop|corner|{}'.format(ma): 5,
                'ontop|corner|{}'.format(mb): 6,
                'bridge|step|{}{}'.format(ma,ma): 7, 
                'bridge|step|{}{}'.format(ma,mb): 8,
                'bridge|step|{}{}'.format(mb,mb): 9,
                'bridge|corner|{}{}'.format(ma,ma): 10,
                'bridge|corner|{}{}'.format(ma,mb): 11,
                'bridge|corner|{}{}'.format(mb,mb): 12,
                'bridge|sc-tc-h|{}{}'.format(ma,ma): 13,
                'bridge|sc-tc-h|{}{}'.format(ma,mb): 14,
                'bridge|sc-tc-h|{}{}'.format(mb,mb): 15,
                # terrace bridge is equivalent to tc-cc-h bridge
                'bridge|tc-cc-h|{}{}'.format(ma,ma): 16,
                'bridge|tc-cc-h|{}{}'.format(ma,mb): 17,
                'bridge|tc-cc-h|{}{}'.format(mb,mb): 18,
                'bridge|sc-cc-t|{}{}'.format(ma,ma): 19,
                'bridge|sc-cc-t|{}{}'.format(ma,mb): 20,
                'bridge|sc-cc-t|{}{}'.format(mb,mb): 21,
                'fcc|sc-tc-h|{}{}{}'.format(ma,ma,ma): 22,
                'fcc|sc-tc-h|{}{}{}'.format(ma,ma,mb): 23, 
                'fcc|sc-tc-h|{}{}{}'.format(ma,mb,mb): 24,
                'fcc|sc-tc-h|{}{}{}'.format(mb,mb,mb): 25,
                'fcc|tc-cc-h|{}{}{}'.format(ma,ma,ma): 26,
                'fcc|tc-cc-h|{}{}{}'.format(ma,ma,mb): 27,
                'fcc|tc-cc-h|{}{}{}'.format(ma,mb,mb): 28,
                'fcc|tc-cc-h|{}{}{}'.format(mb,mb,mb): 29,
                'hcp|sc-tc-h|{}{}{}'.format(ma,ma,ma): 30,
                'hcp|sc-tc-h|{}{}{}'.format(ma,ma,mb): 31,
                'hcp|sc-tc-h|{}{}{}'.format(ma,mb,mb): 32,
                'hcp|sc-tc-h|{}{}{}'.format(mb,mb,mb): 33,
                'hcp|tc-cc-h|{}{}{}'.format(ma,ma,ma): 34,
                'hcp|tc-cc-h|{}{}{}'.format(ma,ma,mb): 35,
                'hcp|tc-cc-h|{}{}{}'.format(ma,mb,mb): 36,
                'hcp|tc-cc-h|{}{}{}'.format(mb,mb,mb): 37,
                '4fold|sc-cc-t|{}{}{}{}'.format(ma,ma,ma,ma): 38,
                '4fold|sc-cc-t|{}{}{}{}'.format(ma,ma,ma,mb): 39, 
                '4fold|sc-cc-t|{}{}{}{}'.format(ma,ma,mb,mb): 40,
                '4fold|sc-cc-t|{}{}{}{}'.format(ma,mb,ma,mb): 41, 
                '4fold|sc-cc-t|{}{}{}{}'.format(ma,mb,mb,mb): 42,
                '4fold|sc-cc-t|{}{}{}{}'.format(mb,mb,mb,mb): 43,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,ma,ma,ma,ma): 44,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,ma,ma,ma,mb): 45,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,ma,mb,ma,ma): 46,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,ma,ma,mb,mb): 47,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,ma,mb,mb,ma): 48,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,ma,mb,mb,mb): 49,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,mb,ma,ma,ma): 50,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,mb,ma,ma,mb): 51,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,mb,mb,ma,ma): 52,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,mb,ma,mb,mb): 53,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,mb,mb,mb,ma): 54,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,mb,mb,mb,mb): 55,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,mb,mb,ma,ma,ma): 56,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,mb,mb,ma,ma,mb): 57,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,mb,mb,mb,ma,ma): 58,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,mb,mb,ma,mb,mb): 59,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,mb,mb,mb,mb,ma): 60,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,mb,mb,mb,mb,mb): 61,
                '6fold|subsurf|{}{}{}{}{}{}'.format(mb,mb,mb,ma,ma,ma): 62,
                '6fold|subsurf|{}{}{}{}{}{}'.format(mb,mb,mb,ma,ma,mb): 63,
                '6fold|subsurf|{}{}{}{}{}{}'.format(mb,mb,mb,mb,ma,ma): 64,
                '6fold|subsurf|{}{}{}{}{}{}'.format(mb,mb,mb,ma,mb,mb): 65,
                '6fold|subsurf|{}{}{}{}{}{}'.format(mb,mb,mb,mb,mb,ma): 66,
                '6fold|subsurf|{}{}{}{}{}{}'.format(mb,mb,mb,mb,mb,mb): 67}
                 
    elif surface in ['fcc311','fcc331']:
        return {'ontop|step|{}'.format(ma): 1,
                'ontop|step|{}'.format(mb): 2,
                'ontop|terrace|{}'.format(ma): 3,
                'ontop|terrace|{}'.format(mb): 4,
                'bridge|step|{}{}'.format(ma,ma): 5,
                'bridge|step|{}{}'.format(ma,mb): 6,
                'bridge|step|{}{}'.format(mb,mb): 7,
                'bridge|terrace|{}{}'.format(ma,ma): 8,
                'bridge|terrace|{}{}'.format(ma,mb): 9,
                'bridge|terrace|{}{}'.format(mb,mb): 10,
                'bridge|sc-tc-h|{}{}'.format(ma,ma): 11,
                'bridge|sc-tc-h|{}{}'.format(ma,mb): 12,
                'bridge|sc-tc-h|{}{}'.format(mb,mb): 13,
                'bridge|sc-tc-t|{}{}'.format(ma,ma): 14,
                'bridge|sc-tc-t|{}{}'.format(ma,mb): 15,
                'bridge|sc-tc-t|{}{}'.format(mb,mb): 16,
                'fcc|sc-tc-h|{}{}{}'.format(ma,ma,ma): 17,
                'fcc|sc-tc-h|{}{}{}'.format(ma,ma,mb): 18,
                'fcc|sc-tc-h|{}{}{}'.format(ma,mb,mb): 19,
                'fcc|sc-tc-h|{}{}{}'.format(mb,mb,mb): 20,
                'hcp|sc-tc-h|{}{}{}'.format(ma,ma,ma): 21,
                'hcp|sc-tc-h|{}{}{}'.format(ma,ma,mb): 22,
                'hcp|sc-tc-h|{}{}{}'.format(ma,mb,mb): 23,
                'hcp|sc-tc-h|{}{}{}'.format(mb,mb,mb): 24,
                '4fold|sc-tc-t|{}{}{}{}'.format(ma,ma,ma,ma): 25,
                '4fold|sc-tc-t|{}{}{}{}'.format(ma,ma,ma,mb): 26, 
                '4fold|sc-tc-t|{}{}{}{}'.format(ma,ma,mb,mb): 27,
                '4fold|sc-tc-t|{}{}{}{}'.format(ma,mb,ma,mb): 28, 
                '4fold|sc-tc-t|{}{}{}{}'.format(ma,mb,mb,mb): 29,
                '4fold|sc-tc-t|{}{}{}{}'.format(mb,mb,mb,mb): 30,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,ma,ma,ma,ma): 31,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,ma,ma,ma,mb): 32,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,ma,mb,ma,ma): 33,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,ma,ma,mb,mb): 34,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,ma,mb,mb,ma): 35,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,ma,mb,mb,mb): 36,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,mb,ma,ma,ma): 37,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,mb,ma,ma,mb): 38,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,mb,mb,ma,ma): 39,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,mb,ma,mb,mb): 40,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,mb,mb,mb,ma): 41,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,mb,mb,mb,mb): 42,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,mb,mb,ma,ma,ma): 43,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,mb,mb,ma,ma,mb): 44,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,mb,mb,mb,ma,ma): 45,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,mb,mb,ma,mb,mb): 46,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,mb,mb,mb,mb,ma): 47,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,mb,mb,mb,mb,mb): 48,
                '6fold|subsurf|{}{}{}{}{}{}'.format(mb,mb,mb,ma,ma,ma): 49,
                '6fold|subsurf|{}{}{}{}{}{}'.format(mb,mb,mb,ma,ma,mb): 50,
                '6fold|subsurf|{}{}{}{}{}{}'.format(mb,mb,mb,mb,ma,ma): 51,
                '6fold|subsurf|{}{}{}{}{}{}'.format(mb,mb,mb,ma,mb,mb): 52,
                '6fold|subsurf|{}{}{}{}{}{}'.format(mb,mb,mb,mb,mb,ma): 53,
                '6fold|subsurf|{}{}{}{}{}{}'.format(mb,mb,mb,mb,mb,mb): 54}

    elif surface == 'fcc322':                                                      
        return {'ontop|step|{}'.format(ma): 1,
                'ontop|step|{}'.format(mb): 2,
                'ontop|terrace|{}'.format(ma): 3,
                'ontop|terrace|{}'.format(mb): 4,
                'ontop|corner|{}'.format(ma): 5,
                'ontop|corner|{}'.format(mb): 6,
                'bridge|step|{}{}'.format(ma,ma): 7, 
                'bridge|step|{}{}'.format(ma,mb): 8,
                'bridge|step|{}{}'.format(mb,mb): 9,
                'bridge|terrace|{}{}'.format(ma,ma): 10,
                'bridge|terrace|{}{}'.format(ma,mb): 11,
                'bridge|terrace|{}{}'.format(mb,mb): 12,
                'bridge|corner|{}{}'.format(ma,ma): 13,
                'bridge|corner|{}{}'.format(ma,mb): 14,
                'bridge|corner|{}{}'.format(mb,mb): 15,
                'bridge|sc-tc-h|{}{}'.format(ma,ma): 16,
                'bridge|sc-tc-h|{}{}'.format(ma,mb): 17,
                'bridge|sc-tc-h|{}{}'.format(mb,mb): 18,                    
                'bridge|tc-cc-h|{}{}'.format(ma,ma): 19,
                'bridge|tc-cc-h|{}{}'.format(ma,mb): 20,
                'bridge|tc-cc-h|{}{}'.format(mb,mb): 21,
                'bridge|sc-cc-t|{}{}'.format(ma,ma): 22,
                'bridge|sc-cc-t|{}{}'.format(ma,mb): 23,
                'bridge|sc-cc-t|{}{}'.format(mb,mb): 24,
                'fcc|terrace|{}{}{}'.format(ma,ma,ma): 25,
                'fcc|terrace|{}{}{}'.format(ma,ma,mb): 26,
                'fcc|terrace|{}{}{}'.format(ma,mb,mb): 27,
                'fcc|terrace|{}{}{}'.format(mb,mb,mb): 28,
                'fcc|sc-tc-h|{}{}{}'.format(ma,ma,ma): 29,
                'fcc|sc-tc-h|{}{}{}'.format(ma,ma,mb): 30, 
                'fcc|sc-tc-h|{}{}{}'.format(ma,mb,mb): 31,
                'fcc|sc-tc-h|{}{}{}'.format(mb,mb,mb): 32,
                'fcc|tc-cc-h|{}{}{}'.format(ma,ma,ma): 33,
                'fcc|tc-cc-h|{}{}{}'.format(ma,ma,mb): 34,
                'fcc|tc-cc-h|{}{}{}'.format(ma,mb,mb): 35,
                'fcc|tc-cc-h|{}{}{}'.format(mb,mb,mb): 36,
                'hcp|terrace|{}{}{}'.format(ma,ma,ma): 37,
                'hcp|terrace|{}{}{}'.format(ma,ma,mb): 38,
                'hcp|terrace|{}{}{}'.format(ma,mb,mb): 39,
                'hcp|terrace|{}{}{}'.format(mb,mb,mb): 40,
                'hcp|sc-tc-h|{}{}{}'.format(ma,ma,ma): 41,
                'hcp|sc-tc-h|{}{}{}'.format(ma,ma,mb): 42,
                'hcp|sc-tc-h|{}{}{}'.format(ma,mb,mb): 43,
                'hcp|sc-tc-h|{}{}{}'.format(mb,mb,mb): 44,
                'hcp|tc-cc-h|{}{}{}'.format(ma,ma,ma): 45,
                'hcp|tc-cc-h|{}{}{}'.format(ma,ma,mb): 46,
                'hcp|tc-cc-h|{}{}{}'.format(ma,mb,mb): 47,
                'hcp|tc-cc-h|{}{}{}'.format(mb,mb,mb): 48,
                '4fold|sc-cc-t|{}{}{}{}'.format(ma,ma,ma,ma): 49,
                '4fold|sc-cc-t|{}{}{}{}'.format(ma,ma,ma,mb): 50, 
                '4fold|sc-cc-t|{}{}{}{}'.format(ma,ma,mb,mb): 51,
                '4fold|sc-cc-t|{}{}{}{}'.format(ma,mb,ma,mb): 52, 
                '4fold|sc-cc-t|{}{}{}{}'.format(ma,mb,mb,mb): 53,
                '4fold|sc-cc-t|{}{}{}{}'.format(mb,mb,mb,mb): 54,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,ma,ma,ma,ma): 55,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,ma,ma,ma,mb): 56,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,ma,mb,ma,ma): 57,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,ma,ma,mb,mb): 58,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,ma,mb,mb,ma): 59,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,ma,mb,mb,mb): 60,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,mb,ma,ma,ma): 61,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,mb,ma,ma,mb): 62,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,mb,mb,ma,ma): 63,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,mb,ma,mb,mb): 64,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,mb,mb,mb,ma): 65,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,mb,mb,mb,mb): 66,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,mb,mb,ma,ma,ma): 67,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,mb,mb,ma,ma,mb): 68,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,mb,mb,mb,ma,ma): 69,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,mb,mb,ma,mb,mb): 70,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,mb,mb,mb,mb,ma): 71,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,mb,mb,mb,mb,mb): 72,
                '6fold|subsurf|{}{}{}{}{}{}'.format(mb,mb,mb,ma,ma,ma): 73,
                '6fold|subsurf|{}{}{}{}{}{}'.format(mb,mb,mb,ma,ma,mb): 74,
                '6fold|subsurf|{}{}{}{}{}{}'.format(mb,mb,mb,mb,ma,ma): 75,
                '6fold|subsurf|{}{}{}{}{}{}'.format(mb,mb,mb,ma,mb,mb): 76,
                '6fold|subsurf|{}{}{}{}{}{}'.format(mb,mb,mb,mb,mb,ma): 77,
                '6fold|subsurf|{}{}{}{}{}{}'.format(mb,mb,mb,mb,mb,mb): 78}

    elif surface in ['fcc221','fcc332']:                                          
        return {'ontop|step|{}'.format(ma): 1,
                'ontop|step|{}'.format(mb): 2,
                'ontop|terrace|{}'.format(ma): 3,
                'ontop|terrace|{}'.format(mb): 4,
                'ontop|corner|{}'.format(ma): 5,
                'ontop|corner|{}'.format(mb): 6,
                'bridge|step|{}{}'.format(ma,ma): 7, 
                'bridge|step|{}{}'.format(ma,mb): 8,
                'bridge|step|{}{}'.format(mb,mb): 9,
                'bridge|terrace|{}{}'.format(ma,ma): 10,
                'bridge|terrace|{}{}'.format(ma,mb): 11,
                'bridge|terrace|{}{}'.format(mb,mb): 12,
                'bridge|corner|{}{}'.format(ma,ma): 13,
                'bridge|corner|{}{}'.format(ma,mb): 14,
                'bridge|corner|{}{}'.format(mb,mb): 15,
                'bridge|sc-tc-h|{}{}'.format(ma,ma): 16,
                'bridge|sc-tc-h|{}{}'.format(ma,mb): 17,
                'bridge|sc-tc-h|{}{}'.format(mb,mb): 18, 
                'bridge|tc-cc-h|{}{}'.format(ma,ma): 19,
                'bridge|tc-cc-h|{}{}'.format(ma,mb): 20,
                'bridge|tc-cc-h|{}{}'.format(mb,mb): 21,
                'bridge|sc-cc-h|{}{}'.format(ma,ma): 22,
                'bridge|sc-cc-h|{}{}'.format(ma,mb): 23,
                'bridge|sc-cc-h|{}{}'.format(mb,mb): 24,
                'fcc|terrace|{}{}{}'.format(ma,ma,ma): 25,
                'fcc|terrace|{}{}{}'.format(ma,ma,mb): 26,
                'fcc|terrace|{}{}{}'.format(ma,mb,mb): 27,
                'fcc|terrace|{}{}{}'.format(mb,mb,mb): 28,
                'fcc|sc-tc-h|{}{}{}'.format(ma,ma,ma): 29,
                'fcc|sc-tc-h|{}{}{}'.format(ma,ma,mb): 30, 
                'fcc|sc-tc-h|{}{}{}'.format(ma,mb,mb): 31,
                'fcc|sc-tc-h|{}{}{}'.format(mb,mb,mb): 32,
                'fcc|tc-cc-h|{}{}{}'.format(ma,ma,ma): 33,
                'fcc|tc-cc-h|{}{}{}'.format(ma,ma,mb): 34,
                'fcc|tc-cc-h|{}{}{}'.format(ma,mb,mb): 35,
                'fcc|tc-cc-h|{}{}{}'.format(mb,mb,mb): 36,
                'fcc|sc-cc-h|{}{}{}'.format(ma,ma,ma): 37,
                'fcc|sc-cc-h|{}{}{}'.format(ma,ma,mb): 38,
                'fcc|sc-cc-h|{}{}{}'.format(ma,mb,mb): 39,
                'fcc|sc-cc-h|{}{}{}'.format(mb,mb,mb): 40,
                'hcp|terrace|{}{}{}'.format(ma,ma,ma): 41,
                'hcp|terrace|{}{}{}'.format(ma,ma,mb): 42,
                'hcp|terrace|{}{}{}'.format(ma,mb,mb): 43,
                'hcp|terrace|{}{}{}'.format(mb,mb,mb): 44,
                'hcp|sc-tc-h|{}{}{}'.format(ma,ma,ma): 45,
                'hcp|sc-tc-h|{}{}{}'.format(ma,ma,mb): 46,
                'hcp|sc-tc-h|{}{}{}'.format(ma,mb,mb): 47,
                'hcp|sc-tc-h|{}{}{}'.format(mb,mb,mb): 48,
                'hcp|tc-cc-h|{}{}{}'.format(ma,ma,ma): 49,
                'hcp|tc-cc-h|{}{}{}'.format(ma,ma,mb): 50,
                'hcp|tc-cc-h|{}{}{}'.format(ma,mb,mb): 51,
                'hcp|tc-cc-h|{}{}{}'.format(mb,mb,mb): 52,
                'hcp|sc-cc-h|{}{}{}'.format(ma,ma,ma): 53,
                'hcp|sc-cc-h|{}{}{}'.format(ma,ma,mb): 54,
                'hcp|sc-cc-h|{}{}{}'.format(ma,mb,mb): 55,
                'hcp|sc-cc-h|{}{}{}'.format(mb,mb,mb): 56,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,ma,ma,ma,ma): 57,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,ma,ma,ma,mb): 58,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,ma,mb,ma,ma): 59,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,ma,ma,mb,mb): 60,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,ma,mb,mb,ma): 61,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,ma,mb,mb,mb): 62,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,mb,ma,ma,ma): 63,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,mb,ma,ma,mb): 64,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,mb,mb,ma,ma): 65,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,mb,ma,mb,mb): 66,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,mb,mb,mb,ma): 67,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,mb,mb,mb,mb): 68,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,mb,mb,ma,ma,ma): 69,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,mb,mb,ma,ma,mb): 70,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,mb,mb,mb,ma,ma): 71,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,mb,mb,ma,mb,mb): 72,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,mb,mb,mb,mb,ma): 73,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,mb,mb,mb,mb,mb): 74,
                '6fold|subsurf|{}{}{}{}{}{}'.format(mb,mb,mb,ma,ma,ma): 75,
                '6fold|subsurf|{}{}{}{}{}{}'.format(mb,mb,mb,ma,ma,mb): 76,
                '6fold|subsurf|{}{}{}{}{}{}'.format(mb,mb,mb,mb,ma,ma): 77,
                '6fold|subsurf|{}{}{}{}{}{}'.format(mb,mb,mb,ma,mb,mb): 78,
                '6fold|subsurf|{}{}{}{}{}{}'.format(mb,mb,mb,mb,mb,ma): 79,
                '6fold|subsurf|{}{}{}{}{}{}'.format(mb,mb,mb,mb,mb,mb): 80}

    elif surface == 'bcc110':
        return {'ontop|terrace|{}'.format(ma): 1, 
                'ontop|terrace|{}'.format(mb): 2,
                'shortbridge|terrace|{}{}'.format(ma,ma): 3,
                'shortbridge|terrace|{}{}'.format(ma,mb): 4,
                'shortbridge|terrace|{}{}'.format(mb,mb): 5,
                'longbridge|terrace|{}{}'.format(ma,ma): 6, 
                'longbridge|terrace|{}{}'.format(ma,mb): 7,
                'longbridge|terrace|{}{}'.format(mb,mb): 8, 
                '3fold|terrace|{}{}{}'.format(ma,ma,ma): 9,
                '3fold|terrace|{}{}{}'.format(ma,ma,mb): 10, 
                '3fold|terrace|{}{}{}'.format(ma,mb,mb): 11,
                '3fold|terrace|{}{}{}'.format(mb,mb,mb): 12}

    elif surface == 'bcc111':                         
        return {'ontop|step|{}'.format(ma): 1,
                'ontop|step|{}'.format(mb): 2,
                'ontop|terrace|{}'.format(ma): 3,
                'ontop|terrace|{}'.format(mb): 4,
                'ontop|corner|{}'.format(ma): 5,
                'ontop|corner|{}'.format(mb): 6,
                'shortbridge|sc-tc-o|{}{}'.format(ma,ma): 7, 
                'shortbridge|sc-tc-o|{}{}'.format(ma,mb): 8,
                'shortbridge|sc-tc-o|{}{}'.format(mb,mb): 9,
                'shortbridge|tc-cc-o|{}{}'.format(ma,ma): 10,
                'shortbridge|tc-cc-o|{}{}'.format(ma,mb): 11,
                'shortbridge|tc-cc-o|{}{}'.format(mb,mb): 12,
                'longbridge|sc-cc-o|{}{}'.format(ma,ma): 13,
                'longbridge|sc-cc-o|{}{}'.format(ma,mb): 14,
                'longbridge|sc-cc-o|{}{}'.format(mb,mb): 15,
                '3fold|sc-tc-cc-o|{}{}{}'.format(ma,ma,ma): 16,
                '3fold|sc-tc-cc-o|{}{}{}'.format(ma,ma,mb): 17, 
                '3fold|sc-tc-cc-o|{}{}{}'.format(ma,mb,mb): 18,
                '3fold|sc-tc-cc-o|{}{}{}'.format(mb,mb,mb): 19}

    elif surface == 'bcc210':
        return {'ontop|step|{}'.format(ma): 1,
                'ontop|step|{}'.format(mb): 2,
                'ontop|terrace|{}'.format(ma): 3,
                'ontop|terrace|{}'.format(mb): 4,
                'ontop|corner|{}'.format(ma): 5,
                'ontop|corner|{}'.format(mb): 6,
                'bridge|step|{}{}'.format(ma,ma): 7, 
                'bridge|step|{}{}'.format(ma,mb): 8,
                'bridge|step|{}{}'.format(mb,mb): 9,
                'bridge|terrace|{}{}'.format(ma,ma): 10,
                'bridge|terrace|{}{}'.format(ma,mb): 11,
                'bridge|terrace|{}{}'.format(mb,mb): 12,
                'bridge|corner|{}{}'.format(ma,ma): 13,
                'bridge|corner|{}{}'.format(ma,mb): 14,
                'bridge|corner|{}{}'.format(mb,mb): 15,
                'bridge|sc-tc-o|{}{}'.format(ma,ma): 16,
                'bridge|sc-tc-o|{}{}'.format(ma,mb): 17,
                'bridge|sc-tc-o|{}{}'.format(mb,mb): 18,
                'bridge|tc-cc-o|{}{}'.format(ma,ma): 19,
                'bridge|tc-cc-o|{}{}'.format(ma,mb): 20,
                'bridge|tc-cc-o|{}{}'.format(mb,mb): 21,
                'bridge|sc-cc-t|{}{}'.format(ma,ma): 22,
                'bridge|sc-cc-t|{}{}'.format(ma,mb): 23,
                'bridge|sc-cc-t|{}{}'.format(mb,mb): 24,
                '3fold|sc-tc-o|{}{}{}'.format(ma,ma,ma): 25,
                '3fold|sc-tc-o|{}{}{}'.format(ma,ma,mb): 26, 
                '3fold|sc-tc-o|{}{}{}'.format(ma,mb,mb): 27,
                '3fold|sc-tc-o|{}{}{}'.format(mb,mb,mb): 28,
                '3fold|tc-cc-o|{}{}{}'.format(ma,ma,ma): 29,
                '3fold|tc-cc-o|{}{}{}'.format(ma,ma,mb): 30,
                '3fold|tc-cc-o|{}{}{}'.format(ma,mb,mb): 31,
                '3fold|tc-cc-o|{}{}{}'.format(mb,mb,mb): 32,
                '4fold|sc-cc-t|{}{}{}{}'.format(ma,ma,ma,ma): 33,
                '4fold|sc-cc-t|{}{}{}{}'.format(ma,ma,ma,mb): 34, 
                '4fold|sc-cc-t|{}{}{}{}'.format(ma,ma,mb,mb): 35,
                '4fold|sc-cc-t|{}{}{}{}'.format(ma,mb,ma,mb): 36, 
                '4fold|sc-cc-t|{}{}{}{}'.format(ma,mb,mb,mb): 37,
                '4fold|sc-cc-t|{}{}{}{}'.format(mb,mb,mb,mb): 38}

    elif surface == 'bcc211':
        return {'ontop|step|{}'.format(ma): 1,
                'ontop|step|{}'.format(mb): 2,
                'bridge|step|{}{}'.format(ma,ma): 3,
                'bridge|step|{}{}'.format(ma,mb): 4,
                'bridge|step|{}{}'.format(mb,mb): 5,
                'bridge|sc-tc-o|{}{}'.format(ma,ma): 6,
                'bridge|sc-tc-o|{}{}'.format(ma,mb): 7,
                'bridge|sc-tc-o|{}{}'.format(mb,mb): 8,
                '3fold|sc-tc-o|{}{}{}'.format(ma,ma,ma): 9,
                '3fold|sc-tc-o|{}{}{}'.format(ma,ma,mb): 10, 
                '3fold|sc-tc-o|{}{}{}'.format(ma,mb,mb): 11,
                '3fold|sc-tc-o|{}{}{}'.format(mb,mb,mb): 12,
                '4fold|terrace|{}{}-{}{}'.format(ma,ma,ma,ma): 13,
                '4fold|terrace|{}{}-{}{}'.format(ma,ma,ma,mb): 14,
                '4fold|terrace|{}{}-{}{}'.format(ma,ma,mb,mb): 15,
                '4fold|terrace|{}{}-{}{}'.format(ma,mb,ma,ma): 16,
                '4fold|terrace|{}{}-{}{}'.format(ma,mb,ma,mb): 17,
                '4fold|terrace|{}{}-{}{}'.format(ma,mb,mb,mb): 18,
                '4fold|terrace|{}{}-{}{}'.format(mb,mb,ma,ma): 19,
                '4fold|terrace|{}{}-{}{}'.format(mb,mb,ma,mb): 20,
                '4fold|terrace|{}{}-{}{}'.format(mb,mb,mb,mb): 21,
                # neighbor elements count clockwise from shorter bond ma
                '5fold|terrace|{}-{}{}{}{}'.format(ma,ma,ma,ma,ma): 22,
                '5fold|terrace|{}-{}{}{}{}'.format(ma,ma,ma,ma,mb): 23,
                '5fold|terrace|{}-{}{}{}{}'.format(ma,ma,ma,mb,mb): 24,
                '5fold|terrace|{}-{}{}{}{}'.format(ma,ma,mb,ma,mb): 25,
                '5fold|terrace|{}-{}{}{}{}'.format(ma,ma,mb,mb,ma): 26,
                '5fold|terrace|{}-{}{}{}{}'.format(ma,ma,mb,mb,mb): 27,
                '5fold|terrace|{}-{}{}{}{}'.format(ma,mb,mb,mb,mb): 28,
                '5fold|terrace|{}-{}{}{}{}'.format(mb,ma,ma,ma,ma): 29,
                '5fold|terrace|{}-{}{}{}{}'.format(mb,ma,ma,ma,mb): 30,
                '5fold|terrace|{}-{}{}{}{}'.format(mb,ma,ma,mb,mb): 31,
                '5fold|terrace|{}-{}{}{}{}'.format(mb,ma,mb,ma,mb): 32,
                '5fold|terrace|{}-{}{}{}{}'.format(mb,ma,mb,mb,ma): 33,
                '5fold|terrace|{}-{}{}{}{}'.format(mb,ma,mb,mb,mb): 34,
                '5fold|terrace|{}-{}{}{}{}'.format(mb,mb,mb,mb,mb): 35}

    elif surface == 'bcc310':
        return {'ontop|step|{}'.format(ma): 1,
                'ontop|step|{}'.format(mb): 2,
                'ontop|terrace|{}'.format(ma): 3,
                'ontop|terrace|{}'.format(mb): 4,
                'bridge|step|{}{}'.format(ma,ma): 5,
                'bridge|step|{}{}'.format(ma,mb): 6,
                'bridge|step|{}{}'.format(mb,mb): 7,
                'bridge|terrace|{}{}'.format(ma,ma): 8,
                'bridge|terrace|{}{}'.format(ma,mb): 9,
                'bridge|terrace|{}{}'.format(mb,mb): 10,
                'bridge|sc-tc-o|{}{}'.format(ma,ma): 11,
                'bridge|sc-tc-o|{}{}'.format(ma,mb): 12,
                'bridge|sc-tc-o|{}{}'.format(mb,mb): 13,
                'bridge|sc-tc-t|{}{}'.format(ma,ma): 14,
                'bridge|sc-tc-t|{}{}'.format(ma,mb): 15,
                'bridge|sc-tc-t|{}{}'.format(mb,mb): 16,
                '3fold|sc-tc-o|{}{}{}'.format(ma,ma,ma): 17,
                '3fold|sc-tc-o|{}{}{}'.format(ma,ma,mb): 18,
                '3fold|sc-tc-o|{}{}{}'.format(ma,mb,mb): 19,
                '3fold|sc-tc-o|{}{}{}'.format(mb,mb,mb): 20,                    
                '4fold|sc-tc-t|{}{}{}{}'.format(ma,ma,ma,ma): 21,
                '4fold|sc-tc-t|{}{}{}{}'.format(ma,ma,ma,mb): 22, 
                '4fold|sc-tc-t|{}{}{}{}'.format(ma,ma,mb,mb): 23,
                '4fold|sc-tc-t|{}{}{}{}'.format(ma,mb,ma,mb): 24, 
                '4fold|sc-tc-t|{}{}{}{}'.format(ma,mb,mb,mb): 25,
                '4fold|sc-tc-t|{}{}{}{}'.format(mb,mb,mb,mb): 26}

    elif surface == 'hcp10m10t':
        return {'ontop|step|{}'.format(ma): 1,
                'ontop|step|{}'.format(mb): 2,
                'ontop|terrace|{}'.format(ma): 3,
                'ontop|terrace|{}'.format(mb): 4,                     
                'bridge|step|{}{}'.format(ma,ma): 5,    
                'bridge|step|{}{}'.format(ma,mb): 6,
                'bridge|step|{}{}'.format(mb,mb): 7,
                'bridge|terrace|{}{}'.format(ma,ma): 8,
                'bridge|terrace|{}{}'.format(ma,mb): 9,
                'bridge|terrace|{}{}'.format(mb,mb): 10,
                'bridge|sc-tc-t|{}{}'.format(ma,ma): 11,
                'bridge|sc-tc-t|{}{}'.format(ma,mb): 12,
                'bridge|sc-tc-t|{}{}'.format(mb,mb): 13,
                # neighbor elements count clockwise from shorter bond ma
                '5fold|subsurf|{}-{}{}{}{}'.format(ma,ma,ma,ma,ma): 14,
                '5fold|subsurf|{}-{}{}{}{}'.format(ma,ma,ma,ma,mb): 15,
                '5fold|subsurf|{}-{}{}{}{}'.format(ma,ma,ma,mb,mb): 16,
                '5fold|subsurf|{}-{}{}{}{}'.format(ma,ma,mb,ma,mb): 17,
                '5fold|subsurf|{}-{}{}{}{}'.format(ma,ma,mb,mb,ma): 18, 
                '5fold|subsurf|{}-{}{}{}{}'.format(ma,ma,mb,mb,mb): 19,
                '5fold|subsurf|{}-{}{}{}{}'.format(ma,mb,mb,mb,mb): 20,
                '5fold|subsurf|{}-{}{}{}{}'.format(mb,ma,ma,ma,ma): 21,
                '5fold|subsurf|{}-{}{}{}{}'.format(mb,ma,ma,ma,mb): 22,
                '5fold|subsurf|{}-{}{}{}{}'.format(mb,ma,ma,mb,mb): 23,
                '5fold|subsurf|{}-{}{}{}{}'.format(mb,ma,mb,ma,mb): 24,
                '5fold|subsurf|{}-{}{}{}{}'.format(mb,ma,mb,mb,ma): 25,
                '5fold|subsurf|{}-{}{}{}{}'.format(mb,ma,mb,mb,mb): 26,
                '5fold|subsurf|{}-{}{}{}{}'.format(mb,mb,mb,mb,mb): 27}

    elif surface == 'hcp10m11':
        return {'ontop|step|{}'.format(ma): 1, 
                'ontop|step|{}'.format(mb): 2,
                'ontop|terrace|{}'.format(ma): 3, 
                'ontop|terrace|{}'.format(mb): 4,                    
                'bridge|step|{}{}'.format(ma,ma): 5, 
                'bridge|step|{}{}'.format(ma,mb): 6,
                'bridge|step|{}{}'.format(mb,mb): 7, 
                'bridge|terrace|{}{}'.format(ma,ma): 8, 
                'bridge|terrace|{}{}'.format(ma,mb): 9,
                'bridge|terrace|{}{}'.format(mb,mb): 10, 
                'bridge|sc-tc-h|{}{}'.format(ma,ma): 11, 
                'bridge|sc-tc-h|{}{}'.format(ma,mb): 12,
                'bridge|sc-tc-h|{}{}'.format(mb,mb): 13, 
                'fcc|sc-tc-h|{}{}{}'.format(ma,ma,ma): 14,
                'fcc|sc-tc-h|{}{}{}'.format(ma,ma,mb): 15, 
                'fcc|sc-tc-h|{}{}{}'.format(ma,mb,mb): 16,
                'fcc|sc-tc-h|{}{}{}'.format(mb,mb,mb): 17,
                'hcp|sc-tc-h|{}{}{}'.format(ma,ma,ma): 18,
                'hcp|sc-tc-h|{}{}{}'.format(ma,ma,mb): 19, 
                'hcp|sc-tc-h|{}{}{}'.format(ma,mb,mb): 20,
                'hcp|sc-tc-h|{}{}{}'.format(mb,mb,mb): 21,
                '4fold|subsurf|{}{}-{}{}'.format(ma,ma,ma,ma): 22,
                '4fold|subsurf|{}{}-{}{}'.format(ma,ma,ma,mb): 23,
                '4fold|subsurf|{}{}-{}{}'.format(ma,ma,mb,mb): 24,
                '4fold|subsurf|{}{}-{}{}'.format(ma,mb,ma,ma): 25,
                '4fold|subsurf|{}{}-{}{}'.format(ma,mb,ma,mb): 26,
                '4fold|subsurf|{}{}-{}{}'.format(ma,mb,mb,mb): 27,
                '4fold|subsurf|{}{}-{}{}'.format(mb,mb,ma,ma): 28,
                '4fold|subsurf|{}{}-{}{}'.format(mb,mb,ma,mb): 29,
                '4fold|subsurf|{}{}-{}{}'.format(mb,mb,mb,mb): 30,
                # neighbor elements count clockwise from shorter bond ma
                '5fold|subsurf|{}-{}{}{}{}'.format(ma,ma,ma,ma,ma): 31,
                '5fold|subsurf|{}-{}{}{}{}'.format(ma,ma,ma,ma,mb): 32,
                '5fold|subsurf|{}-{}{}{}{}'.format(ma,ma,ma,mb,mb): 33,
                '5fold|subsurf|{}-{}{}{}{}'.format(ma,ma,mb,ma,mb): 34,
                '5fold|subsurf|{}-{}{}{}{}'.format(ma,ma,mb,mb,ma): 35, 
                '5fold|subsurf|{}-{}{}{}{}'.format(ma,ma,mb,mb,mb): 36,
                '5fold|subsurf|{}-{}{}{}{}'.format(ma,mb,mb,mb,mb): 37,
                '5fold|subsurf|{}-{}{}{}{}'.format(mb,ma,ma,ma,ma): 38,
                '5fold|subsurf|{}-{}{}{}{}'.format(mb,ma,ma,ma,mb): 39,
                '5fold|subsurf|{}-{}{}{}{}'.format(mb,ma,ma,mb,mb): 40,
                '5fold|subsurf|{}-{}{}{}{}'.format(mb,ma,mb,ma,mb): 41,
                '5fold|subsurf|{}-{}{}{}{}'.format(mb,ma,mb,mb,ma): 42,
                '5fold|subsurf|{}-{}{}{}{}'.format(mb,ma,mb,mb,mb): 43,
                '5fold|subsurf|{}-{}{}{}{}'.format(mb,mb,mb,mb,mb): 44,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,ma,ma,ma,ma): 45,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,ma,ma,ma,mb): 46,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,ma,mb,ma,ma): 47,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,ma,ma,mb,mb): 48,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,ma,mb,mb,ma): 49,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,ma,mb,mb,mb): 50,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,mb,ma,ma,ma): 51,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,mb,ma,ma,mb): 52,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,mb,mb,ma,ma): 53,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,mb,ma,mb,mb): 54,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,mb,mb,mb,ma): 55,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,mb,mb,mb,mb): 56,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,mb,mb,ma,ma,ma): 57,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,mb,mb,ma,ma,mb): 58,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,mb,mb,mb,ma,ma): 59,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,mb,mb,ma,mb,mb): 60,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,mb,mb,mb,mb,ma): 61,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,mb,mb,mb,mb,mb): 62,
                '6fold|subsurf|{}{}{}{}{}{}'.format(mb,mb,mb,ma,ma,ma): 63,
                '6fold|subsurf|{}{}{}{}{}{}'.format(mb,mb,mb,ma,ma,mb): 64,
                '6fold|subsurf|{}{}{}{}{}{}'.format(mb,mb,mb,mb,ma,ma): 65,
                '6fold|subsurf|{}{}{}{}{}{}'.format(mb,mb,mb,ma,mb,mb): 66,
                '6fold|subsurf|{}{}{}{}{}{}'.format(mb,mb,mb,mb,mb,ma): 67,
                '6fold|subsurf|{}{}{}{}{}{}'.format(mb,mb,mb,mb,mb,mb): 68}

    elif surface == 'hcp10m12':
        return {'ontop|step|{}'.format(ma): 1,
                'ontop|step|{}'.format(mb): 2,
                'ontop|terrace|{}'.format(ma): 3,
                'ontop|terrace|{}'.format(mb): 4,
                'ontop|corner|{}'.format(ma): 5,
                'ontop|corner|{}'.format(mb): 6,
                'bridge|step|{}{}'.format(ma,ma): 7, 
                'bridge|step|{}{}'.format(ma,mb): 8,
                'bridge|step|{}{}'.format(mb,mb): 9,
                'bridge|terrace|{}{}'.format(ma,ma): 10,
                'bridge|terrace|{}{}'.format(ma,mb): 11,
                'bridge|terrace|{}{}'.format(mb,mb): 12,
                'bridge|corner|{}{}'.format(ma,ma): 13,
                'bridge|corner|{}{}'.format(ma,mb): 14,
                'bridge|corner|{}{}'.format(mb,mb): 15,
                'bridge|sc-tc-h|{}{}'.format(ma,ma): 16,
                'bridge|sc-tc-h|{}{}'.format(ma,mb): 17,
                'bridge|sc-tc-h|{}{}'.format(mb,mb): 18,
                'bridge|tc-cc-t|{}{}'.format(ma,ma): 19,
                'bridge|tc-cc-t|{}{}'.format(ma,mb): 20,
                'bridge|tc-cc-t|{}{}'.format(mb,mb): 21,
                'bridge|sc-cc-h|{}{}'.format(ma,ma): 22,
                'bridge|sc-cc-h|{}{}'.format(ma,mb): 23,
                'bridge|sc-cc-h|{}{}'.format(mb,mb): 24,
                'fcc|sc-tc-h|{}{}{}'.format(ma,ma,ma): 25,
                'fcc|sc-tc-h|{}{}{}'.format(ma,ma,mb): 26, 
                'fcc|sc-tc-h|{}{}{}'.format(ma,mb,mb): 27,
                'fcc|sc-tc-h|{}{}{}'.format(mb,mb,mb): 28,
                'fcc|sc-cc-h|{}{}{}'.format(ma,ma,ma): 29,
                'fcc|sc-cc-h|{}{}{}'.format(ma,ma,mb): 30,
                'fcc|sc-cc-h|{}{}{}'.format(ma,mb,mb): 31,
                'fcc|sc-cc-h|{}{}{}'.format(mb,mb,mb): 32,
                'hcp|sc-tc-h|{}{}{}'.format(ma,ma,ma): 33,
                'hcp|sc-tc-h|{}{}{}'.format(ma,ma,mb): 34,
                'hcp|sc-tc-h|{}{}{}'.format(ma,mb,mb): 35,
                'hcp|sc-tc-h|{}{}{}'.format(mb,mb,mb): 36,
                'hcp|sc-cc-h|{}{}{}'.format(ma,ma,ma): 37,
                'hcp|sc-cc-h|{}{}{}'.format(ma,ma,mb): 38,
                'hcp|sc-cc-h|{}{}{}'.format(ma,mb,mb): 39,
                'hcp|sc-cc-h|{}{}{}'.format(mb,mb,mb): 40,
                '4fold|tc-cc-t|{}{}{}{}'.format(ma,ma,ma,ma): 41,
                '4fold|tc-cc-t|{}{}{}{}'.format(ma,ma,ma,mb): 42, 
                '4fold|tc-cc-t|{}{}{}{}'.format(ma,ma,mb,mb): 43,
                '4fold|tc-cc-t|{}{}{}{}'.format(ma,mb,ma,mb): 44, 
                '4fold|tc-cc-t|{}{}{}{}'.format(ma,mb,mb,mb): 45,
                '4fold|tc-cc-t|{}{}{}{}'.format(mb,mb,mb,mb): 46,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,ma,ma,ma,ma): 47,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,ma,ma,ma,mb): 48,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,ma,mb,ma,ma): 49,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,ma,ma,mb,mb): 50,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,ma,mb,mb,ma): 51,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,ma,mb,mb,mb): 52,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,mb,ma,ma,ma): 53,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,mb,ma,ma,mb): 54,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,mb,mb,ma,ma): 55,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,mb,ma,mb,mb): 56,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,mb,mb,mb,ma): 57,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,ma,mb,mb,mb,mb): 58,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,mb,mb,ma,ma,ma): 59,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,mb,mb,ma,ma,mb): 60,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,mb,mb,mb,ma,ma): 61,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,mb,mb,ma,mb,mb): 62,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,mb,mb,mb,mb,ma): 63,
                '6fold|subsurf|{}{}{}{}{}{}'.format(ma,mb,mb,mb,mb,mb): 64,
                '6fold|subsurf|{}{}{}{}{}{}'.format(mb,mb,mb,ma,ma,ma): 65,
                '6fold|subsurf|{}{}{}{}{}{}'.format(mb,mb,mb,ma,ma,mb): 66,
                '6fold|subsurf|{}{}{}{}{}{}'.format(mb,mb,mb,mb,ma,ma): 67,
                '6fold|subsurf|{}{}{}{}{}{}'.format(mb,mb,mb,ma,mb,mb): 68,
                '6fold|subsurf|{}{}{}{}{}{}'.format(mb,mb,mb,mb,mb,ma): 69,
                '6fold|subsurf|{}{}{}{}{}{}'.format(mb,mb,mb,mb,mb,mb): 70}


# Use this label dictionary when site compostion is 
# considered. Useful for slabs with arbitrary number
# of components.
def get_multimetallic_slab_labels(surface, metals): 
    if surface in ['fcc111','hcp0001']:
        relations = {'ontop': ['terrace'],
                     'bridge': ['terrace'],
                     'fcc': ['terrace'],
                     'hcp': ['terrace'],
                     '6fold': ['subsurf']}

    elif surface in ['fcc100','bcc100']:
        relations = {'ontop': ['terrace'],
                     'bridge': ['terrace'],
                     '4fold': ['terrace']}

    elif surface in ['fcc110','hcp10m10h']:
        relations = {'ontop': ['step'], 
                     'bridge': ['step','sc-tc-h'],
                     'fcc': ['sc-tc-h'],
                     'hcp': ['sc-tc-h'],
                     '4fold': ['terrace'],
                     '5fold': ['terrace'],
                     '6fold': ['subsurf']}

    elif surface == 'fcc211':
        relations = {'ontop': ['step','terrace','corner'],
                     'bridge': ['step','terrace','corner','sc-tc-h','tc-cc-h','sc-cc-t'],
                     'fcc': ['sc-tc-h','tc-cc-h'],
                     'hcp': ['sc-tc-h','tc-cc-h'],
                     '4fold': ['sc-cc-t'],
                     '6fold': ['subsurf']}

    elif surface in ['fcc311','fcc331']:
        relations = {'ontop': ['step','terrace'],
                     'bridge': ['step','terrace','sc-tc-h','sc-tc-t'],
                     'fcc': ['sc-tc-h'],
                     'hcp': ['sc-tc-h'],
                     '4fold': ['sc-tc-t'],
                     '6fold': ['subsurf']}

    elif surface == 'fcc322':
        relations = {'ontop': ['step','terrace','corner'],
                     'bridge': ['step','terrace','corner','sc-tc-h','tc-cc-h','sc-cc-t'],
                     'fcc': ['terrace','sc-tc-h','tc-cc-h'],
                     'hcp': ['terrace','sc-tc-h','tc-cc-h'],
                     '4fold': ['sc-cc-t'],
                     '6fold': ['subsurf']}

    elif surface in ['fcc221','fcc332']:
        relations = {'ontop': ['step','terrace','corner'],
                     'bridge': ['step','terrace','corner','sc-tc-h','tc-cc-h','sc-cc-h'],
                     'fcc': ['terrace','sc-tc-h','tc-cc-h'],
                     'hcp': ['terrace','sc-tc-h','tc-cc-h'],
                     '6fold': ['subsurf']}

    elif surface == 'bcc110':
        relations = {'ontop': ['terrace'],
                     'shortbridge': ['terrace'],
                     'longbridge': ['terrace'],
                     '3fold': ['terrace']}

    elif surface == 'bcc111':
        relations = {'ontop': ['step','terrace','corner'],
                     'shortbridge': ['sc-tc-o','tc-cc-o'],
                     'longbridge': ['sc-cc-o'],
                     '3fold': ['sc-tc-cc-o']}

    elif surface == 'bcc210':
        relations = {'ontop': ['step','terrace','corner'],
                     'bridge': ['step','terrace','corner','sc-tc-o','tc-cc-o','sc-cc-t'],
                     '3fold': ['sc-tc-o','tc-cc-o'],
                     '4fold': ['sc-cc-t']}

    elif surface == 'bcc211':
        relations = {'ontop': ['step'], 
                     'bridge': ['step','sc-tc-o'],
                     '3fold': ['sc-tc-o'],
                     '4fold': ['terrace'],
                     '5fold': ['terrace']}

    elif surface == 'bcc310':
        relations = {'ontop': ['step','terrace'],
                     'bridge': ['step','terrace','sc-tc-o','sc-tc-t'],
                     '3fold': ['sc-tc-o'],
                     '4fold': ['sc-tc-t']}

    elif surface == 'hcp10m10t':
        relations = {'ontop': ['step','terrace'],
                     'bridge': ['step','terrace','sc-tc-t'],
                     '5fold': ['subsurf']}

    elif surface == 'hcp10m11':
        relations = {'ontop': ['step','terrace'],
                     'bridge': ['step','terrace','sc-tc-h'],
                     'fcc': ['sc-tc-h'],
                     'hcp': ['sc-tc-h'],
                     '4fold': ['subsurf'],
                     '5fold': ['subsurf'],
                     '6fold': ['subsurf']}

    elif surface == 'hcp10m12':
        relations = {'ontop': ['step','terrace','corner'],
                     'bridge': ['step','terrace','corner','sc-tc-h','tc-cc-t','sc-cc-h'],
                     'fcc': ['sc-tc-h','sc-cc-h'],
                     'hcp': ['sc-tc-h','sc-cc-h'],
                     '4fold': ['tc-cc-t'],
                     '6fold': ['subsurf']}

    d = {}
    count = 0 
    for k, vs in relations.items():
        if k == 'ontop':
            n = 1
        elif k in ['bridge','shortbridge','longbridge']:
            n = 2
        elif k in ['fcc','hcp','3fold']:
            n = 3
        elif k == '4fold':
            n = 4
        elif k == '5fold':
            n = 5
        elif k == '6fold':
            n = 6
        for v in vs:
            for p in product(metals, repeat=n):
                if k == '5fold':
                    comp = str(p[0]) + '-' + ''.join(p[1:])
                elif k == '4fold' and surface in ['fcc110','hcp10m10h','bcc211','hcp10m11']:
                    comp = ''.join(p[:2]) + '-' + ''.join(p[2:])
                else: 
                    comp = ''.join(p)
                count += 1
                d['{0}|{1}|{2}'.format(k, v, comp)] = count 
                
    return d


def get_cluster_signature_from_label(label, composition_effect=False, metals=[]):
    if not composition_effect:
        label_dict = get_monometallic_cluster_labels()
    else:
        if len(metals) <= 2:
            label_dict = get_bimetallic_cluster_labels(metals)
        else:
            label_dict = get_multimetallic_cluser_labels(metals)

    return list(label_dict.keys())[list(label_dict.values()).index(int(label))]


def get_slab_signature_from_label(label, surface, composition_effect=False, metals=[]):
    if not composition_effect:
        label_dict = get_monometallic_slab_labels(surface)
    else:
        if len(metals) <= 2:
            label_dict = get_bimetallic_slab_labels(surface, metals)
        else:
            label_dict = get_multimetallic_slab_labels(surface, metals)

    return list(label_dict.keys())[list(label_dict.values()).index(int(label))]



from acat.adsorbate_coverage import SlabAdsorbateCoverage
from acat.utilities import (neighbor_shell_list, 
                            get_adj_matrix, 
                            hash_composition)
from multiprocessing import Pool
from itertools import chain, combinations
from functools import partial
import networkx as nx
import numpy as np
import os


class WL(object):

    def __init__(self,
                 hp={'length': np.array([.75, .25])},
                 normalize=False,
                 atom_wise=True,
                 dx=0.5,
                 connect_nn=False,
                 n_jobs=os.cpu_count(),
                 **kwargs):
        self.hp = hp
        self.set_hyperparams(hp)
        self.normalize = normalize
        self.atom_wise = atom_wise
        self.dx = dx
        self.connect_nn = connect_nn
        self.n_jobs = n_jobs
        self.kwargs = kwargs

    def set_hyperparams(self, new_params):
        """Set or update the hyperparameters for the Kernel.

        Parameters
        ----------
        new_params: dictionary
            A dictionary of hyperparameters that are added or updated.
        """
        self.hp.update(new_params)
#        if 'alpha' not in self.hp:
#            self.hp['alpha'] = np.array([.75])
        if 'length' not in self.hp:
            self.hp['length'] = np.array([200.])

        # Lower and upper machine precision
        eps_mach_lower = np.sqrt(1.01 * np.finfo(float).eps)
        eps_mach_upper = 1 / eps_mach_lower
        self.hp['length'] = np.abs(self.hp['length']).reshape(-1)
#        self.hp['alpha'] = np.abs(self.hp['alpha']).reshape(-1)
        self.hp['length'] = np.where(self.hp['length'] < eps_mach_upper, 
                                     np.where(self.hp['length'] > eps_mach_lower, 
                                              self.hp['length'], eps_mach_lower), 
                                     eps_mach_upper)
#        self.hp['alpha'] = np.where(self.hp['alpha'] < eps_mach_upper, 
#                                    np.where(self.hp['alpha'] > eps_mach_lower, 
#                                             self.hp['alpha'], eps_mach_lower), 
#                                    eps_mach_upper)
        return self.hp

    def dist_m(self, train_images, test_images=None):
        dists = self.__call__(train_images, test_images, get_dists=True)

        return dists

    def __call__(self, train_images, test_images=None, get_derivatives=False, 
                 get_dists=False, dists=None):
        if test_images is None:
            images = list(train_images)
        else:
            images = list(train_images) + list(test_images)
        if dists is None:                                      
            pool = Pool(self.n_jobs)
            dicts = pool.map(self.get_dict, images)
#            for d in dicts:
#                all_keys = set(chain.from_iterable(dicts))   
#                old_keys = d.keys()
#                d.update((k, 0) for k in all_keys - old_keys)
            dicts = np.asarray(dicts, dtype=object)
            if get_dists:
                t = len(self.hp['length'])                                                  
                Kdts = [np.zeros(shape=(len(dicts), len(dicts)))] * t
                idx0dt, idx1dt = np.triu_indices_from(Kdts[0])
                for idt in range(len(idx0dt)):
                    i0, i1 = idx0dt[idt], idx1dt[idt]
                    for k in set(dicts[i0]).intersection(set(dicts[i1])):
                        maxrep = count = 0
                        current = ''
                        for c in k:
                            if c == current == ':':
                                count += 1
                            else:
                                count = 1
                                current = c
                            maxrep = max(count, maxrep)
                        Kdts[maxrep-1][i0,i1] += 1
                for j, Kdt in enumerate(Kdts):                   
                    Kdts[j] = Kdt + Kdt.T - np.diag(np.diag(Kdt))
                Kdts = np.asarray(Kdts)
                if t == 1:
                    Kdts = Kdts[0]
                dists = (dicts, Kdts)
                return dists
        else:
            dicts = dists[0]
            print('length: {}'.format(self.hp['length']))
#            print('alpha: {}'.format(self.hp['alpha']))
            print('noise: {}'.format(self.hp['noise']))

        # Initialize the similarity kernel
        K = np.zeros(shape=(len(images), len(images)))
        # Iterate through upper triangular matrix indices
        idx0, idx1 = np.triu_indices_from(K)
        # Perform similarity calculations in parallel.
        pool = Pool(self.n_jobs)
        po = partial(self.pairwise_operation)
        sims = pool.starmap(po, zip(dicts[idx0], dicts[idx1]))

        # Insert it into K[i,j] and K[j,i]
        for i, sim in enumerate(sims):
            K[idx0[i],idx1[i]] = sim
        # Symmetrize
        K = K + K.T - np.diag(np.diag(K))
        # Normalize
        if self.normalize:
            d = np.diag(K)**-0.5
            K = np.einsum("i,ij,j->ij", d, K, d)
        # Take the upper right rectangle if there are test images
        if test_images is not None:
            K = K[:len(train_images),len(train_images):len(images)] 

        return K

    def get_hyperparameters(self):
        "Get all the hyperparameters"
        return self.hp 

    def get_dict(self, atoms):
        d = {} 
        numbers = atoms.numbers
        if self.atom_wise:
            nblist = neighbor_shell_list(atoms, dx=self.dx, neighbor_number=1, 
                                         mic=(True in atoms.pbc))                      
            A = get_adj_matrix(nblist)                                                     
        else:            
            sac = SlabAdsorbateCoverage(atoms, **self.kwargs)
            A = sac.get_graph(atom_wise=False, return_adj_matrix=True, 
                              full_effect=True, connect_dentates=True)                  
        G = nx.from_numpy_matrix(A)

        t = len(self.hp['length'])
        nnlabs, neighbors, lpd = {}, {}, {}
        isolates = []
        for i in range(len(A)):
            lab0 = str(numbers[i])
            if lab0 in d:
                d[lab0] += 1.
            else:
                d[lab0] = 1.

            if t > 0:
                nnd = nx.single_source_shortest_path_length(G, i, cutoff=1)
                nns = [j for j, v in nnd.items() if v == 1]
                neighbors[i] = nns
                if len(nns) == 0:
                    isolates.append(i)
                    continue
                if self.connect_nn:
                    nnhood = np.asarray([i] + nns)
                    An = A[nnhood,:][:,nnhood]
                    Gn = nx.from_numpy_matrix(An)
                    # An algorithm to find the lexicographically minimum 
                    # longest path starts from node and ends at a neighbor
                    maxlen = 0
                    longest_paths = []
  
                    for ani, anj in combinations(range(len(An)), 2):
                        paths = nx.all_simple_paths(Gn, source=ani, target=anj)
                        for path in paths:
                            length = len(path)
                            if length > maxlen:
                                longest_paths = [path]
                                maxlen = length
                            elif length == maxlen:
                                longest_paths.append(path)
                    mint = min(((idx, hash_composition(numbers[nnhood[lp]])) for idx, lp 
                                in enumerate(longest_paths)), key=lambda x: x[1])
                    longest_path = longest_paths[mint[0]]
                    sorted_numbers = mint[1]
                    lpd[i] = longest_path
                else:
                    sorted_numbers = sorted(numbers[nns])
                lab1 = lab0 + ':' + ','.join(map(str, sorted_numbers))
                nnlabs[i] = lab1
                if lab1 in d:
                    d[lab1] += self.hp['length'][0] 
                else:
                    d[lab1] = self.hp['length'][0]
#        print(sorted(d.keys()))

        if t > 1: 
            for k in range(1, t):
                nnnlabs = {}
                for i in range(len(A)):
                    if i in isolates:
                        continue
                    nnlab = nnlabs[i]
                    if self.connect_nn:
                        nnnlab = ','.join([nnlabs[nn] for nn in lpd[i]])
                    else:
                        nnnlab = ','.join(sorted(nnlabs[nn] for nn in neighbors[i]))
                    lab = nnlab + ':' * (k + 1) + nnnlab
                    nnnlabs[i] = lab
                    if lab in d:
                        d[lab] += self.hp['length'][k]
                    else:
                        d[lab] = self.hp['length'][k]
#                    print('Atom{}: '.format(i))
#                    dic = {'28:28,28,28,28': '3', '28:8,28,28,28,28': '4', 
#                           '28:8,8,28,28,28,28': '5', '8:28,28,28,28': '6'}
#                    for ki, kj in dic.items():
#                        lab = lab.replace(ki, kj)
#                        print(lab)
                nnlabs = nnnlabs 

        return d

    def get_derivatives(self, train_images, hp, KXX=None, dists=None):                 
        """Get the derivatives of the kernel matrix in respect to the 
        hyperparameters.

        Parameters
        ----------
        train_images : list of ase.Atoms
            N training images.
        hp : list
            A list with hyperparameters to calculate derivatives.
        KXX : (N,N) array (optional)
            The kernel matrix of training data.
        dists : (N,N) array (optional)
            Can be given the distance matrix to avoid recaulcating it. 
        """
        assert not self.normalize
        hp_deriv = {}
        if 'length' in hp:
            if dists is None:                                      
                pool = Pool(self.n_jobs)
                dicts = pool.map(self.get_dict, train_images)
                dicts = np.asarray(dicts, dtype=object)
                t = len(self.hp['length'])                                                   
                Kdts = [np.zeros(shape=(len(dicts), len(dicts)))] * t
                idx0dt, idx1dt = np.triu_indices_from(Kdt)
                for idt in range(len(idx0dt)):
                    i0, i1 = idx0dt[idt], idx1dt[idt]
                    for k in set(dicts[i0]).intersection(set(dicts[i1])):
                        maxrep = count = 0
                        current = ''
                        for c in k:
                            if c == current == ':':
                                count += 1
                            else:
                                count = 1
                                current = c
                            maxrep = max(count, maxrep)
                        Kdts[maxrep-1][i0,i1] += 1
                for j, Kdt in enumerate(Kdts):                   
                    Kdts[j] = Kdt + Kdt.T - np.diag(np.diag(Kdt))
                Kdts = np.asarray(Kdts)
                if t == 1:
                    Kdts = Kdts[0]
            else:
                Kdts = dists[1]
        hp_deriv['length'] = Kdts

        return hp_deriv

    def pairwise_operation(self, dx, dy):
        """Compute pairwise similarity between every two dicts."""

        dxy = {k: [d[k] if k in d else 0 for d in (dx, dy)] 
               for k in set(dx.keys()) | set(dy.keys())}
        X = np.asarray(list(dxy.values()))
 
       # Inner product
        k = X[:,0] @ X[:,1].T

        # Exponential
#        k = np.exp(-np.linalg.norm(X[:,0] - X[:,1]) / self.hp['length'][0]**2)

        # Square exponential
#        k = np.exp(-np.sum((X[:,0] - X[:,1])**2) / (2 * self.hp['length'][0]**2))

        return k

    def diag(self, test_images, get_derivatives=False):
        """Get the diagonal kernel vector.

        Parameters
        ----------
        test_images : (N,N) array
            Test images with N structures.
        """

        if self.normalize:
            return np.ones(len(test_images))

        K = self.__call__(train_images=test_images)
        return np.diag(K)

    def __repr__(self):
        return 'WL(hp={})'.format(self.hp)
